import typing

from .. import Device, Undefined, VISADevice, util
from .. import paramattr as attr

if typing.TYPE_CHECKING:
    import numpy as np
    import pandas as pd
else:
    pd = util.lazy_import('pandas')
    np = util.lazy_import('numpy')

__all__ = ['PowerSensor', 'Oscilloscope', 'SignalGenerator']


@attr.visa_keying(
    # the default SCPI query and write formats
    query_fmt='{key}?',
    write_fmt='{key} {value}',
    # map python True and False values to these SCPI strings
    remap={True: 'ON', False: 'OFF'},
)
# set the automatic connection filters
class PowerSensor(VISADevice):
    make = attr.value.str('FakeTech', inherit=True)
    model = attr.value.str('Power Sensor #1234', inherit=True)
    write_termination: str = attr.value.str('\r\n', inherit=True)

    RATES = 'NORM', 'DOUB', 'FAST'

    # SCPI string keys and bounds on the parameter values,
    # taken from the instrument programming manual
    initiate_continuous = attr.property.bool(
        key='INIT:CONT', help='trigger continuously if True'
    )
    trigger_count = attr.property.int(
        key='TRIG:COUN', help='acquisition count', label='samples', min=1, max=200
    )
    measurement_rate = attr.property.str(
        key='SENS:MRAT',
        only=RATES,
        case=False,
    )
    sweep_aperture = attr.property.float(
        key='SWE:APER', help='measurement duration', label='s', min=20e-6, max=200e-3
    )
    frequency = attr.property.float(
        key='SENS:FREQ',
        help='calibration frequency',
        label='Hz',
        min=10e6,
        max=18e9,
        step=1e-3,
    )

    def preset(self):
        """revert to instrument default presets"""
        self.write('SYST:PRES')

    def fetch(self):
        """acquire measurements as configured"""
        response = self.query('FETC?')

        if self.trigger_count == 1:
            return float(response)
        else:
            return pd.Series([float(s) for s in response.split(',')], name='spectrum')

    def fetch_ascii_values(self):
        """acquire measurements as configured"""
        s = self.query_ascii_values('FETC?', container=pd.Series)
        s.name = 'spectrum'
        return s

    def trigger(self):
        return self.write('TRIG')


@attr.visa_keying(remap={True: 'ON', False: 'OFF'})
class SpectrumAnalyzer(VISADevice):
    make = attr.value.str('FakeTech', inherit=True)
    model = attr.value.str('Spectrum Analyzer #1234', inherit=True)

    center_frequency = attr.property.float(
        key='SENS:FREQ',
        min=10e6,
        max=18e9,
        step=1e-3,
        help='input signal center frequency',
        label='Hz',
    )
    resolution_bandwidth = attr.property.float(
        key='SENS:BW',
        min=1,
        max=40e6,
        step=1e-3,
        help='resolution bandwidth',
        label='Hz',
    )

    def load_state(self, remote_filename: str):
        """revert to instrument preset state"""
        self.write(f"LOAD '{remote_filename}'")

    def fetch(self):
        """acquire measurements as configured"""
        response = self.query('FETC?')

        series = pd.Series(
            [float(s) for s in response.split(',')], name='power_spectral_density'
        )
        series.index = pd.Index(
            self.center_frequency + np.linspace(-5e6, 5e6, len(series)),
            name='frequency',
        )

        return series

    def fetch_ascii_values(self):
        """acquire measurements as configured"""
        s = self.query_ascii_values('FETC?', container=pd.Series)
        s.name = 'spectrum'
        return s

    def trigger(self):
        util.sleep(0.1)  # simulate slow response
        return self.write('TRIG')


@attr.visa_keying(remap={True: 'YES', False: 'NO'})
class SignalGenerator(VISADevice):
    make = attr.value.str('FakeTech', inherit=True)
    model = attr.value.str('Signal Generator #1234', inherit=True)

    output_enabled = attr.property.bool(
        key='OUT:ENABL', help='when True, output an RF tone'
    )
    center_frequency = attr.property.float(
        key='SENS:FREQ',
        help='input signal center frequency',
        label='Hz',
        min=10e6,
        max=18e9,
        step=1e-3,
    )
    mode = attr.property.str(key='MODE', only=['sweep', 'tone', 'iq'], case=False)

    def trigger(self):
        """revert to instrument preset state"""
        self.write('TRIG')


@attr.method_kwarg.int('channel', min=1, max=4, help='input channel')
@attr.visa_keying(remap={True: 'ON', False: 'OFF'})
class Oscilloscope(VISADevice):
    make = attr.value.str('FakeTech', inherit=True)
    model = attr.value.str('Oscilloscope #1234', inherit=True)

    @attr.method.float(
        label='Hz',
        help='channel center frequency',
        min=10e6,
        max=18e9,
        step=1e-3,
    )
    def center_frequency(self, set_value=Undefined, /, *, channel):
        if set_value is Undefined:
            return self.query(f'CH{channel}:SENS:FREQ?')
        else:
            self.write(f'CH{channel}:SENS:FREQ {set_value}')

    resolution_bandwidth = attr.method.float(
        key='CH{channel}:SENS:BW',
        help='channel resolution bandwidth',
        label='Hz',
        min=1,
        max=40e6,
        step=1e-3,
        # arguments omitted deliberately for testing
    )

    def load_state(self, remote_filename: str):
        """revert to instrument preset state"""
        self.write(f"LOAD '{remote_filename}'")

    def fetch(self):
        """acquire measurements as configured"""
        response = self.query('FETC?')

        series = pd.Series([float(s) for s in response.split(',')], name='spectrum')
        series.index = pd.Index(
            self.center_frequency + np.linspace(-5e6, 5e6, len(series)),
            name='frequency',
        )

        return series

    def trigger(self):
        return self.write('TRIG')


class LocalStoreDevice(Device):
    def open(self):
        self.backend = {}



from . import _types
from ._bases import Value


class any(Value, _types.Any):
    """defines a value attribute that validates any python object"""


class bool(Value, _types.Bool):
    """defines a value attribute that casts to python builtin bool"""


class float(Value, _types.Float):
    """defines a value attribute that casts to python builtin float"""


class int(Value, _types.Int):
    """defines a value attribute that casts to python builtin int"""


class complex(Value, _types.Complex):
    """defines a value attribute that casts to python complex"""


class str(Value, _types.Unicode):
    """defines a value attribute that casts to python builtin str"""


class bytes(Value, _types.Bytes):
    """defines a value attribute that casts to python builtin bytes"""


class list(Value, _types.List):
    """defines a value attribute that casts to python builtin list"""


class tuple(Value, _types.Tuple):
    """defines a value attribute that casts to python builtin tuple"""


class dict(Value, _types.Dict):
    """defines a value attribute that casts to python builtin dict"""


class Path(Value, _types.Path):
    """defines a value attribute that casts to python pathlib.Path"""


class NetworkAddress(Value, _types.NetworkAddress):
    """defines a value attribute that casts to a network address"""


# used to set up hints for static type checking
_ALL_TYPES = (
    Value,
    any,
    bool,
    float,
    int,
    complex,
    str,
    bytes,
    list,
    tuple,
    dict,
    Path,
    NetworkAddress,
)



import copy
import inspect
import io
import os
import pickle
import re
import warnings
from collections.abc import Callable, Iterable
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Any, TypeVar, TYPE_CHECKING, Union

from . import _device, _host, _rack, util
from . import _device as core
from . import paramattr as attr
from ._device import Device
from ._rack import Owner
from .paramattr import observe

if TYPE_CHECKING:
    # not executed: help static code analysis recognize lazy_imports
    import json
    import numpy as np
    import pandas as pd
    import sqlalchemy
    import pyarrow
    import shutil
    import tarfile
else:
    json = util.lazy_import('json')
    np = util.lazy_import('numpy')
    pd = util.lazy_import('pandas')
    sqlalchemy = util.lazy_import('sqlalchemy')
    pyarrow = util.lazy_import('pyarrow')
    shutil = util.lazy_import('shutil')
    tarfile = util.lazy_import('tarfile')

# define this alias type to avoid a forced import
DataFrameType = TypeVar('DataFrameType', bound='pd.DataFrame')

EMPTY = inspect._empty

INSPECT_SKIP_FILES = _device.__file__, attr.__file__, _rack.__file__, __file__


class MungerBase(core.Device):
    """Organize file output with a key in the root database.

    The following conversions to relational files are attempted in order to
    convert each value in the row dictionary:

    1. Text containing a valid file or directory *outside* of the root data
       directory is made relational by moving the file or directory into
       the current row. The text is replaced with the updated relative path;
    2. Text longer than `text_relational_min` is dumped into a relational
       text file;
    3. 1- or 2-D data is converted to a pandas Series or DataFrame, and
       dumped into a relational file defined by the extension set by
       `nonscalar_file_type`

    """

    resource: Path = attr.value.Path(
        allow_none=True, kw_only=False, help='base directory for all data'
    )
    text_relational_min: int = attr.value.int(
        default=1024,
        min=0,
        help='minimum size threshold that triggers storing text in a relational file',
    )
    force_relational: list = attr.value.list(
        default=[], help='list of column names to always save as relational data'
    )
    relational_name_fmt: str = attr.value.str(
        default='{id}',
        help='directory name format for data in each row keyed on column',
    )
    nonscalar_file_type: str = attr.value.str(
        default='csv', help='file format for non-scalar numerical data'
    )
    metadata_dirname = 'metadata'

    def __call__(self, index, row):
        """
        Break special cases of row items that need to be stored in
        relational files. The row valueis replaced in-place with the relative
        path to the data saved on disk.

        Arguments:
            row: dictionary of {'entry_name': "entry_value"} pairs
        Returns:
            the row dictionary, replacing special entries with the relative path to the saved data file
        """

        def is_path(v):
            if not isinstance(v, str):
                return False
            try:
                return os.path.exists(v)
            except ValueError:
                return False

        for name, v in row.items():
            if is_path(v):
                # Path to a datafile to move into the dataset
                row[name] = self._from_external_file(name, v, index, row)

            elif isinstance(v, (str, bytes)):
                # A long string that should be written to a text file
                if (
                    len(v) > self.text_relational_min
                    or name in self.force_relational
                    or isinstance(v, bytes)
                ):
                    row[name] = self._from_text(name, v, index, row)

            elif isinstance(v, (np.ndarray, pd.Series, pd.DataFrame)):
                # vector, table, matrix, etc.
                row[name] = self._from_ndarraylike(name, v, index, row)

            elif hasattr(v, '__len__') or hasattr(v, '__iter__'):
                # tuple, list, or other iterable
                row[name] = self._from_sequence(name, v, index, row)

        return row

    def __repr__(self):
        return f"{type(self).__name__}('{self.resource!s}')"

    def get_metadata(
        self,
        name,
        name_func: Callable[[attr.HasParamAttrs, str, dict], str],
        **metadata,
    ):
        def process_value(value, key_name):
            if isinstance(value, (str, bytes)):
                if len(value) > self.text_relational_min:
                    self._from_text(key_name, value)
                else:
                    return value
            elif isinstance(value, (pd.DataFrame, np.ndarray)):
                self._from_ndarraylike(key_name, value)
            else:
                return value

        global_values = {}
        for owner, owner_name in name.items():
            if not isinstance(owner, Device):
                # other util.Ownable instances e.g. RelationalTableLogger
                continue

            for attr_name, attr_def in attr.get_class_attrs(owner).items():
                if isinstance(attr_def, attr.value.Value) or attr_def.cache:
                    summary_key = name_func(owner, attr_name)
                    global_values[summary_key] = getattr(owner, attr_name)
        global_values = {k: process_value(v, k) for k, v in global_values.items()}

        # metadata = dict(summary=pd.DataFrame([summary], index=["Value"]).T)
        metadata['global_values'] = global_values

        return metadata

    def save_metadata(self, metadata):
        raise NotImplementedError

    def _from_ndarraylike(self, name, value, index=0, row=None):
        """Write nonscalar (potentially array-like, or a python object) data
        to a file, and return a path to the file

        Arguments:
            name: name of the entry to write, used as the filename
            value: the object containing array-like data
            row: row dictionary, or None (the default) to write to the metadata folder
        Returns:
            the path to the file, relative to the directory that contains the root database
        """

        def write(stream, ext, value):
            if ext == 'csv':
                value.to_csv(stream)
            elif ext == 'json':
                value.to_json(stream)
            elif ext in ('p', 'pickle'):
                pickle.dump(value, stream, 2)
            elif ext == 'feather':
                to_feather(value, stream)
            elif ext == 'db':
                raise Exception('sqlite not implemented for relational files')
            else:
                raise Exception(f"extension {ext} doesn't match a known format")

        if hasattr(value, '__len__') and len(value) == 0:
            return ''

        if row is None:
            ext = 'csv'
        else:
            ext = self.nonscalar_file_type

        try:
            value = pd.DataFrame(value)
            if value.shape[0] == 0:
                value = pd.DataFrame([value])
        except BaseException:
            # We couldn't make a DataFrame
            self._logger.error(
                f'Failed to form DataFrame from {name!r}; pickling object instead'
            )
            ext = 'pickle'
        finally:
            if row is None:
                stream = self._open_metadata(name + '.' + ext, 'wb')
            else:
                stream = self._open_relational(name + '.' + ext, index, row, mode='wb')

            # Workaround for bytes/str encoding quirk underlying pandas 0.23.1
            try:
                write(stream, ext, value)
                self._get_key(stream)
            except TypeError:
                with io.TextIOWrapper(stream, newline='\n') as buf:
                    write(buf, ext, value)
            finally:
                stream.close()
        return self._get_key(stream)

    def _from_external_file(self, name, old_path, index=0, row=None, ntries=10):
        basename = os.path.basename(old_path)
        with self._open_relational(basename, index, row, mode='wb') as buf:
            new_path = buf.name

        self._relational_from_file(old_path, new_path)

        return self._get_key(new_path)

    def _from_text(self, name, value, index=0, row=None, ext='.txt'):
        """Write a string data to a file

        Arguments:
            name: name of the parameter (helps to determine file path)
            value: the string to write to file
            row: the row to infer timestamp, or None to write to metadata
            ext: file extension
        Returns:
            the path to the file, relative to the directory that contains the root database
        """
        with self._open_relational(name + ext, index, row, 'w') as f:
            f.write(value)
        return self._get_key(f)

    def _from_sequence(self, name, value, index=0, row=None, ext='.json'):
        """Write a string data to a file

        Arguments:
            name: name of the parameter (helps to determine file path)
            value: the string to write to file
            row: the row to infer timestamp, or None to write to metadata
            ext: file extension
        Returns:
            the path to the file, relative to the directory that contains the root database
        """
        with self._open_relational(name + ext, index, row, mode='w') as fd:
            json.dump(value, fd, indent=True)  # f.write(bytes(value, encoding='utf-8'))

        return self._get_key(fd)

    # The following methods need to be implemented in subclasses.
    def _get_key(self, buf):
        """Key to use for the relative data in the root database?

        Arguments:
            stream: stream for writing to the relational data file
        Returns:
            the key
        """
        raise NotImplementedError

    def _open_relational(self, name, index, row, mode):
        """Open a stream / IO buffer for writing relational data, given
        the root database column, index, and the row dictionary.

        Arguments:
            name: the column name of the relational data in the root db
            index: the index name of of the data in the root db
            row: the dictionary containing the row of data at `index`

        Returns:

            an open buffer object for writing data
        """
        raise NotImplementedError

    def _open_metadata(self, name, mode, in_root=False):
        """Open a stream / IO buffer for writing metadata, given
        the name of the metadata.

        Returns:

            an open buffer object for writing metadata to the file
        """

        raise NotImplementedError

    def _relational_from_file(self, old_path, dest):
        raise NotImplementedError

    def format_metadata_value(self, key_name, value):
        if isinstance(value, (str, bytes)):
            if len(value) > self.text_relational_min:
                self._from_text(key_name, value)
            else:
                return value
        elif isinstance(value, (pd.DataFrame, np.ndarray)):
            self._from_ndarraylike(key_name, value)
        else:
            return value

    def open(self):
        # touch these modules to ensure they have imported
        pd.DataFrame
        np.array


class MungeToDirectory(MungerBase):
    """Implement data munging into subdirectories."""

    def _open_relational(self, name, index, row, mode):
        relpath = self._make_path_heirarchy(index, row)
        if not os.path.exists(relpath):
            os.makedirs(relpath)

        return open(os.path.join(relpath, name), mode)

    def _open_metadata(self, name, mode, in_root=False):
        if in_root:
            dirpath = self.resource
        else:
            dirpath = self.resource / self.metadata_dirname
            dirpath.mkdir(exist_ok=True)

        return open(Path(dirpath) / name, mode)

    def _get_key(self, stream):
        """Key to use for the relative data in the root database?

        Arguments:
            stream: stream for writing to the relational data file
        Returns:
            the key
        """
        return os.path.relpath(stream.name, self.resource)

    def _relational_from_file(self, old_path, dest):
        # Retry moves for several seconds in case the file is in the process
        # of being closed
        @util.until_timeout(PermissionError, 5, delay=0.5)
        def move():
            os.rename(old_path, dest)

        try:
            move()
            self._logger.debug(f'moved {old_path!r} to {dest!r}')
        except PermissionError:
            self._logger.warning(
                'relational file was still open in another program; fallback to copy instead of rename'
            )
            shutil.copyfile(old_path, dest)

    def _make_path_heirarchy(self, index: int, row: dict):
        # TODO: add back in a timestamp
        relpath = self.relational_name_fmt.format(id=index, **row)

        # TODO: More robust enforcement for a valid path strings
        relpath = relpath.replace(':', '')
        relpath = os.path.join(self.resource, relpath)
        return relpath

    def save_metadata(self, metadata: dict[str, Any]):
        def recursive_dict_fix(d):
            d = dict(d)
            for name, obj in dict(d).items():
                if isinstance(obj, Path):
                    d[name] = str(obj)
                elif isinstance(obj, bytes):
                    d[name] = obj.decode()
                elif isinstance(obj, dict):
                    d[name] = recursive_dict_fix(obj)
            return d

        sanitized = recursive_dict_fix(metadata)

        # for k, v in metadata.items():
        #     stream = self._open_metadata(k + ".json", "w")
        #     if hasattr(stream, "overwrite"):
        #         stream.overwrite = True

        #     if isinstance(v, pd.DataFrame):
        #         v = v.to_dict()#["Value"]
        #     if isinstance(v, dict):
        #         v = recursive_dict_fix(v)

        with self._open_metadata('metadata.json', 'w', in_root=True) as stream:
            json.dump(sanitized, stream, indent=True)


class TarFileIO(io.BytesIO):
    """For appending data into new files in a tarfile"""

    def __init__(self, open_tarfile, relname, mode='w', overwrite=False):
        self.tarfile = open_tarfile
        self.overwrite = False
        self.name = relname
        self.mode = mode
        super(TarFileIO, self).__init__()

    def __del__(self):
        try:
            super(TarFileIO, self).close()
        except ValueError:
            pass

        super(TarFileIO, self).__del__()

    def write(self, data, encoding='ascii'):
        if isinstance(data, str):
            data = bytes(data, encoding=encoding)
        super(TarFileIO, self).write(data)

    def close(self):
        # First: dump the data into the tar file
        try:
            if not self.overwrite and self.name in self.tarfile.getnames():
                raise OSError(f'{self.name} already exists in {self.tarfile.name}')

            tarinfo = tarfile.TarInfo(self.name)
            tarinfo.size = self.tell()

            self.seek(0)

            self.tarfile.addfile(tarinfo, self)

        # Then make sure to close everything
        finally:
            super(TarFileIO, self).close()


class MungeToTar(MungerBase):
    """Implement data munging into a tar file. This is slower than
    MungeToDirectory but is tidier on the filesystem.
    """

    tarname = 'data.tar'

    def _open_relational(self, name, index, row, mode):
        directory = self.relational_name_fmt.format(id=index, **row)
        relpath = f'{directory}/{name}'
        return TarFileIO(self.tarfile, relpath, mode=mode)

    def _open_metadata(self, name, mode, in_root=False):
        if in_root:
            return MungeToDirectory._open_metadata(self, name, mode, True)
        if in_root:
            dirpath = name
        else:
            dirpath = os.path.join(self.metadata_dirname, name)
        stream = TarFileIO(self.tarfile, dirpath, mode=mode)
        if hasattr(stream, 'overwrite'):
            stream.overwrite = True
        return stream

    def open(self):
        if not os.path.exists(self.resource):
            with suppress(FileExistsError):
                os.makedirs(self.resource)
        self.tarfile = tarfile.open(os.path.join(self.resource, self.tarname), 'a')

    def close(self):
        self.tarfile.close()

    def _get_key(self, buf):
        """Where is the file relative to the root database?

        Arguments:
            path: path of the file
        Returns:
            path to the file relative to the root database
        """
        return buf.name

    def _relational_from_file(self, old_path, dest):
        info = self.tar.gettarinfo(old_path)
        info.name = dest
        self.tar.addfile(old_path, info)

        @util.until_timeout(PermissionError, 5, delay=0.5)
        def remove():
            try:
                shutil.rmtree(old_path)
            except NotADirectoryError:
                os.remove(old_path)

        try:
            remove()
            self._logger.debug(f'moved {old_path} to into tar file as {dest}')
        except PermissionError:
            self._logger.warning(f'could not remove old file or directory {old_path}')

    def save_metadata(self, metadata: dict[str, Any]):
        def recursive_dict_fix(d):
            d = dict(d)
            for name, obj in dict(d).items():
                if isinstance(obj, Path):
                    d[name] = str(obj)
                elif isinstance(obj, bytes):
                    d[name] = obj.decode()
                elif isinstance(obj, dict):
                    d[name] = recursive_dict_fix(obj)
            return d

        sanitized = recursive_dict_fix(metadata)

        # for k, v in metadata.items():
        #     stream = self._open_metadata(k + ".json", "w")
        #     if hasattr(stream, "overwrite"):
        #         stream.overwrite = True

        #     if isinstance(v, pd.DataFrame):
        #         v = v.to_dict()#["Value"]
        #     if isinstance(v, dict):
        #         v = recursive_dict_fix(v)

        with self._open_metadata('metadata.json', 'w', in_root=True) as stream:
            json.dump(sanitized, stream, indent=True)


class Aggregator(util.Ownable):
    """Manages aggregation of parameters of Device attributes defined with paramattr, and data returned by calls to methods in Rack instances"""

    # note: this is not a dataclass, these annotations are only for type-hinting
    name_map: dict[attr.HasParamAttrs, str]
    attr_rules: dict[str, dict[attr.HasParamAttrs, str]]

    # observed value changes from paramattr objects
    incoming_attr_always: dict[str, attr.HasParamAttrs]
    incoming_attr_auto: dict[str, attr.HasParamAttrs]

    def __init__(self):
        # registry of names to use for HasParamAttr subclasses like Device
        self.name_map = {}
        self.attr_rules = dict(always={}, never={})

        # pending data
        self.incoming_attr_auto = {}
        self.incoming_attr_always = {}
        self.incoming_rack_output = {}
        self.incoming_rack_input = {}

        self._rack_toplevel_caller = None
        self._rack_input_index = 0

        # self._iter_index_names = {}

        # cached data
        self.metadata = {
            'device_objects': {},
            'global_values': {},
            'field_name_sources': {},
        }

        super().__init__()

    def __repr__(self):
        return f'{self.__class__.__qualname__}()'

    def enable(self):
        # catch return data as well
        _rack.notify.observe_returns(self._receive_rack_output)
        _rack.notify.observe_calls(self._receive_rack_input)
        # _rack.notify.observe_call_iteration(self._receive_rack_iteration)

    def disable(self):
        _rack.notify.unobserve_returns(self._receive_rack_output)
        _rack.notify.unobserve_calls(self._receive_rack_input)
        # _rack.notify.unobserve_call_iteration(self._receive_rack_input)

    def always_get_this_attr(self, device, attr_name):
        if not isinstance(device, core.Device):
            return False

        attr_def = attr.get_class_attrs(device)[attr_name]

        return isinstance(attr_def, attr.value.Value) or attr_def.cache

    def get(self) -> list([dict, dict]):
        """return an aggregated dictionary output data (from Device paramattr objects and Rack method returns)
        and an aggregated dictionary of input data (keyword arguments passed to the toplevel Rack method).
        A get is
        also performed on each Device trait that is configured as "always" with `self.observe`, and any paramattr objects
        labeled "never" are removed.

        Returns:
            dictionary keyed on :func:`key` (defaults '{device name}_{attr name}_{repr of method kwargs}')
        """

        for device in self.name_map.keys():
            # Perform gets for each property trait called out in self.trait_rules['always']
            if device in self.attr_rules['always'].keys():
                for attr_name in self.attr_rules['always'][device]:
                    field_name = self.name_attr_field(device, attr_name)
                    value = getattr(device, attr_name)

                    if self.always_get_this_attr(device, attr_name):
                        self.incoming_attr_always[field_name] = value
                    else:
                        self.incoming_attr_auto[field_name] = value

            # Remove keys corresponding with self.trait_rules['never']
            if device in self.attr_rules['never'].keys():
                for attr_name in self.attr_rules['never'][device]:
                    field_name = self.name_attr_field(device, attr_name)
                    if self.always_get_this_attr(device, attr_name):
                        self.incoming_attr_always.pop(field_name, None)
                    else:
                        self.incoming_attr_auto.pop(field_name, None)

        aggregated_output = {}

        # start by aggregating the trait data, and checking for conflicts with keys in the Rack data
        aggregated_output.update(self.incoming_attr_always)
        aggregated_output.update(self.incoming_attr_auto)

        # check namespace conflicts and pull in rack outputs
        key_conflicts = set(aggregated_output).intersection(self.incoming_rack_output)
        if len(key_conflicts) > 0:
            self.critical(
                f'key name conflict in aggregated data - Rack data is overwriting trait data for {key_conflicts}'
            )
        aggregated_output.update(self.incoming_rack_output)

        # and the rack inputs
        aggregated_input = {}
        aggregated_input.update(self.incoming_rack_input)

        if 'index' in aggregated_input:
            aggregated_output['index'] = aggregated_input['index']

        # clear Rack data, as well as property trait data if we don't assume it is consistent.
        # value traits are locally cached, so it is safe to keep them in the next step
        self.incoming_rack_output = {}
        self.incoming_attr_auto = {}
        # self._pending_rack_input = {}

        return aggregated_output, aggregated_input

    def get_metadata(self, process_func: callable):
        metadata = dict(self.metadata)
        global_values = {}
        for owner, owner_name in self.name_map.items():
            if not isinstance(owner, Device):
                # other util.Ownable instances e.g. RelationalTableLogger
                continue

            for attr_name, attr_def in attr.get_class_attrs(owner).items():
                if isinstance(attr_def, attr.value.Value) or attr_def.cache:
                    summary_key = self.name_attr_field(owner, attr_name)
                    global_values[summary_key] = getattr(owner, attr_name)
        global_values = {k: process_func(v, k) for k, v in global_values.items()}

        metadata['global_values'] = global_values

        return metadata

    def sanitize_column_name(self, string: str) -> str:
        return re.sub(r'\W+|^(?=\d)+', '_', string)

    def name_attr_field(
        self, device: attr.HasParamAttrs, attr_name: str, kwargs: dict[str, Any] = {}
    ):
        """Generate a name for a trait based on the names of
        a device and one of its states or paramattr.
        """
        kwarg_repr = ', '.join([f'{k}={v!r}' for k, v in kwargs.items()])
        attr_def = getattr(type(device), attr_name)

        bare_name = f'{self.name_map[device]}_{attr_name}_{kwarg_repr}'
        sanitized_name = self.sanitize_column_name(bare_name).rstrip('_')

        if attr_def._type.__module__ != 'builtins':
            typename = f'{attr_def._type.__module__}.{attr_def._type.__qualname__}'
        else:
            typename = attr_def._type.__qualname__

        attr_metadata = {
            'object': f'{self.name_map[device]}.{attr_name}',
            'paramattr': attr_def.__repr__(),
            'type': typename,
            'help': attr_def.help,
            'label': attr_def.label,
        }

        if isinstance(attr_def, attr.method.Method):
            attr_metadata['kwargs'] = kwargs

        self.metadata['field_name_sources'][sanitized_name] = attr_metadata

        return sanitized_name

    def set_device_labels(self, **mapping: dict[Device, str]):
        """Manually choose device name for a device instance.

        Arguments:
            mapping: name mapping, formatted as {device_object: 'device name'}
        Returns:
            None
        """
        for label, device in list(mapping.items()):
            if isinstance(device, Aggregator):
                pass
            elif not isinstance(device, Device):
                raise ValueError(f'{device} is not an instance of Device')

        self.name_map.update([(v, k) for k, v in mapping.items()])

    def _receive_rack_output(self, msg: dict):
        """called by an owning Rack notifying that managed procedural steps have returned data"""
        # trait data or previous returned data may cause problems here. perhaps this should be an exception?

        row_data = msg['new']

        key_conflicts = set(row_data).intersection(self.incoming_rack_output)
        if len(key_conflicts) > 0:
            self._logger.warning(
                f'Rack call overwrites prior data with existing keys {key_conflicts}'
            )
        self.incoming_rack_output.update(row_data)

    def _receive_rack_input(self, msg: dict):
        """called by an owning Rack notifying that managed procedural steps have returned data"""
        # trait data or previous returned data may cause problems here. perhaps this should be an exception?

        row_data = dict(index=self._rack_input_index)

        if self._rack_toplevel_caller is None:
            # take the first call as the top-level caller
            self._rack_toplevel_caller = msg['owner']

        elif self._rack_toplevel_caller != msg['owner']:
            # only track calls to the top-level caller
            return

        else:
            # iter_info = msg["new"]
            self._rack_input_index += 1

        # if len(self._iter_index_names) > 0:
        #     self._iter_index_names.setdefault(
        #         msg["owner"], "index_" + msg["owner"]._owned_name
        #     )
        # else:
        #     self._iter_index_names.setdefault(msg["owner"], "index")

        # row_data[self._iter_index_names[msg["owner"]]] = iter_info["index"]
        # if iter_info['step_name']:
        #     row_data[msg['owner']._owned_name + '_step_name'] = iter_info['step_name']

        # TODO: should there be some kind of conflict check for this?
        # key_conflicts = set(row_data).intersection(self._pending_traits_persistent)
        # if len(key_conflicts) > 0:
        #     self._logger.warning(
        #         f"Rack call overwrites prior data with existing keys {key_conflicts}"
        #     )

        self.incoming_rack_input = dict(row_data, **msg['new'])

    def _receive_paramattr_update(self, msg: dict):
        """called by trait owners on changes observed

        Arguments:
            msg: callback info dictionary generated by traitlets
        Returns:
            None
        """

        if msg['name'].startswith('_'):
            # skip private traits
            return

        if not msg['paramattr'].log:
            return

        name = self.name_map[msg['owner']]
        attr_name = msg['name']
        kwargs = msg.get('kwargs', {})
        data_name = self.name_attr_field(msg['owner'], attr_name, kwargs)

        if msg['cache']:
            self.metadata[data_name] = msg['new']
        elif self.always_get_this_attr(msg['owner'], msg['name']):
            # TODO: need to figure out the semantics around "always" attrs that are methods
            self.incoming_attr_always[data_name] = msg['new']
        elif not name.startswith('_'):
            self.incoming_attr_auto[data_name] = msg['new']

    def update_name_map(
        self,
        ownables: dict[Device, str],
        owner_prefix: Union[str, None] = None,
        fallback_names={},
    ):
        """map each Device to a name in devices.values() by introspection."""
        if owner_prefix is None:
            prefix = ''
        else:
            prefix = owner_prefix + '.'

        for obj, name in ownables.items():
            if not isinstance(obj, util.Ownable):
                raise ValueError(
                    f'{obj} is not a Device or other ownable labbench type'
                )

            if name is not None:
                if obj not in self.name_map:
                    self.name_map[obj] = new_name = prefix + name
                else:
                    new_name = name

            elif obj in self.name_map:
                # naming for device. needs this
                new_name = prefix + self.name_map[obj]

            elif getattr(obj, '_owned_name', None) is not None:
                self.name_map[obj] = new_name = obj._owned_name

            elif hasattr(obj, '__name__'):
                # Ownable instances that have owners have this attribute set
                self.name_map[obj] = new_name = prefix + obj.__name__

            else:
                try:
                    self.name_map[obj] = new_name = prefix + self.inspect_object_name(
                        obj
                    )
                    self._logger.info(f"{obj} named '{new_name}' by introspection")
                except RuntimeError:
                    if obj in fallback_names:
                        self.name_map[obj] = new_name = prefix + fallback_names[obj]
                        self._logger.info(
                            f"{obj} named '{new_name}' based on fallback after introspection failed"
                        )
                    else:
                        raise

            if isinstance(obj, Owner):
                children = dict(zip(obj._ownables.values(), obj._ownables.keys()))
                self.update_name_map(children, owner_prefix=new_name)

            self.metadata['device_objects'][new_name] = repr(obj)

        if len(list(self.name_map.values())) != len(set(self.name_map.values())):
            duplicates = {}
            for k, v in self.name_map.items():
                duplicates.setdefault(v, []).append(k)

            duplicates = {k: v for k, v in duplicates.items() if len(v) > 1}

            raise RuntimeError(
                f'could not automatically resolve duplicate device name(s) {duplicates}'
            )

    def observe(
        self,
        devices,
        changes: bool = True,
        always: Union[str, list[str]] = [],
        never: Union[str, list[str]] = ['isopen'],
    ):
        """Configure the data to aggregate from value, property, or datareturn traits in the given devices.

        Device may be a single device instance, or an
        several devices in an iterable (such as a list
        or tuple) to apply to each one.

        Subsequent calls to :func:`observe_states` replace
        the existing list of observed property traits for each
        device.

        Arguments:
            devices: Device instance, iterable of Devices, or {device:name} mapping
            changes: Whether to automatically log each time a property trait is set for the supplied device(s)
            always: name (or iterable of multiple names) of property traits to actively update on each call to get()
            never: name (or iterable of multiple names) of property traits to exclude from aggregated result (overrides :param:`always`)
        """

        # parameter checks
        if isinstance(devices, Device):
            devices = {devices: None}
        elif isinstance(devices, dict):
            pass
        elif hasattr(devices, '__iter__'):
            devices = dict([(d, None) for d in devices])
        else:
            raise ValueError('devices argument must be a device or iterable of devices')

        if isinstance(always, str):
            always = (always,)
        elif hasattr(always, '__iter__'):
            always = tuple(always)
        else:
            raise ValueError("argument 'always' must be a str or iterable of str")

        if isinstance(never, str):
            never = (never,)
        elif hasattr(never, '__iter__'):
            never = tuple(never)
        else:
            raise ValueError("argument 'never' must be a str or iterable of str")

        self.update_name_map(devices)

        # Register handlers
        for device in devices.keys():
            if changes:
                observe(device, self._receive_paramattr_update)
            else:
                attr.unobserve(device, self._receive_paramattr_update)
            if always:
                self.attr_rules['always'][device] = always
            if never:
                self.attr_rules['never'][device] = never

    def inspect_object_name(self, target, max_levels=20):
        """Introspect into the caller to name an object .

        Arguments:
            target: device instance
            min_levels: minimum number of layers of calls to traverse
            max_levels: maximum number of layers to traverse
        Returns:
            name of the Device
        """

        def find_value(haystack, needle, reject=['self']):
            for k, v in haystack.items():
                if v is needle:
                    if k in reject:
                        raise KeyError
                    else:
                        return k
            else:
                raise KeyError

        f = frame = inspect.currentframe()

        ret = None

        try:
            while f.f_code.co_filename in INSPECT_SKIP_FILES:
                f = f.f_back

            tries_elapsed = 0
            while tries_elapsed < max_levels:
                # look in the namespace; if nothing
                try:
                    ret = find_value(f.f_locals, target)
                except KeyError:
                    f = f.f_back
                    if f is None:
                        tries_elapsed = max_levels
                    else:
                        tries_elapsed += 1
                    continue
                else:
                    break
            else:
                raise RuntimeError(
                    f'failed to automatically label {target!r} by inspection'
                )
        finally:
            del f, frame

        return ret


class ParamAttrLogger(
    Owner, util.Ownable, entry_order=(_host.Email, MungerBase, _host.Host)
):
    """Base class for loggers that queue dictionaries of data before writing
    to disk. This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path: Root path to use for the root database
        overwrite: Whether to overwrite the root database if it exists (otherwise, append)

        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data

        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
        git_commit_in: perform a git commit on open() if the current
        directory is inside a git repo with this branch name
    """

    INDEX_LABEL = 'id'

    def __init__(
        self,
        path: Path,
        *,
        append: bool = False,
        text_relational_min: int = 1024,
        force_relational: list[str] = ['host_log'],
        nonscalar_file_type: str = 'csv',
        tar: bool = False,
    ):
        self.path = Path(path)
        self.last_row = 0
        self.pending_output = []
        self.pending_input = []
        self._append = append

        self.aggregator = Aggregator()

        # log host introspection
        # TODO: smarter

        self.host = _host.Host()

        # select the backend that dumps relational data
        munge_cls = MungeToTar if tar else MungeToDirectory
        self.munge = munge_cls(
            path,
            text_relational_min=text_relational_min,
            force_relational=force_relational,
            nonscalar_file_type=nonscalar_file_type,
            # **metadata
        )

        self.set_row_preprocessor(None)

        # doing this at the end ensures self._logger is seeded with appropriate log messages
        super().__init__()

    def __copy__(self):
        return copy.deepcopy(self)

    def __owner_init__(self, owner: Owner):
        """observe children of each Rack or other Owner"""
        super().__owner_init__(owner)
        devices, _ = _rack.recursive_devices(owner)
        self.observe_paramattr({v: k for k, v in devices.items()})

    def __repr__(self):
        if hasattr(self, 'path'):
            path = repr(str(self.path))
        else:
            path = '<unset path>'
        return f'{self.__class__.__qualname__}({path})'

    def observe_paramattr(
        self,
        devices: Union[Device, Iterable[Device]],
        changes: bool = True,
        always: Union[str, Iterable[str]] = [],
        never: Union[str, Iterable[str]] = ['isopen'],
    ):
        """Configure aggregation of activity in parameter attributes defined with `labbench.paramattr.value`,
        `labbench.paramattr.property`, or labbench.paramattr.method`.

        Device may be a single device instance or an iterable several devices in an iterable (such as a list
        or tuple) to apply to each one.

        Subsequent calls to :func:`observe_states` replace
        the existing list of observed property traits for each
        device.

        Arguments:
            devices: One or more Device instances to observe
            changes: Whether to automatically log each time a property trait is set for the supplied device(s)
            always: name (or iterable of multiple names) of property traits to actively update on each call to get()
            never: name (or iterable of multiple names) of property traits to exclude from aggregated result (overrides :param:`always`)
        """

        self.aggregator.observe(
            devices=devices, changes=changes, always=always, never=never
        )

    def set_row_preprocessor(self, func):
        """Define a function that is called to modify each pending data row
        before it is committed to disk. It should accept a single argument,
        a function or other callable that accepts a single argument (the
        row dictionary) and returns the dictionary modified for write
        to disk.
        """
        if func is None:
            self._row_preprocessor = lambda x: x
        elif callable(func):
            self._row_preprocessor = func
        else:
            raise ValueError('func argument must be callable or None')

    def new_row(self, *args, **kwargs):
        """Add a new row of data to the list of rows that await writes
        to disk, `self.pending_output`. This cache of pending data row is in
        the dictionary `self.pending_output`.

        The data row is represented as a dict in the form {'column_name': value}.
        This dict is formed here with
        * the most recent value, property, and datareturn traits ({trait_name: latest_value} as configured by `self.observe`);
        * optional positional dict args[0] (each {key: value});
        * optional keyword arguments in `kwargs` (each {key: value})

        Keyword arguments are passed in as::

            db.new_row(dict(name0=value0), name1=value1, name2=value2, nameN=valueN)

        Simple "scalar" database types like numbers, booleans, and strings
        are added directly to the table. Non-scalar or multidimensional values are stored
        in a separate file (as defined in :func:`set_path_format`), and the path
        to this file is stored in the table.

        In order to write `self.pending_output` to disk, use :func:`self.write`.

        :param bool copy=True: When `True` (the default), use a deep copy of `data` to avoid
        problems with overwriting references to data if `data` is reused during test. This takes some extra time; set to `False` to skip this copy operation.

        Returns:
            the dictionary representation of the row added to `self.pending_output`.
        """
        do_copy = kwargs.pop('copy', None)

        # Start with input arguments
        if len(args) == 1:
            #            if not isinstance(args[0], dict):
            #                raise TypeError('argument to append must be a dictionary')
            row = copy.deepcopy(dict(args[0])) if do_copy else dict(args[0])
        elif len(args) == 0:
            row = {}

        # Pull in observed states. aggregated_input comes from Rack method keyword args
        aggregated_output, aggregated_input = self.aggregator.get()
        if do_copy:
            aggregated_output = copy.deepcopy(aggregated_output)
        row.update(aggregated_output)

        # Pull in keyword arguments
        if do_copy:
            kwargs = copy.deepcopy(dict(kwargs))
        else:
            kwargs = dict(kwargs)
        row.update(kwargs)

        self._logger.debug(f'new data row has {len(row)} columns')

        self.pending_output.append(row)
        self.pending_input.append(aggregated_input)

    def write(self):
        """Commit any pending rows to the root database, converting
        non-scalar data to data files, and replacing their dictionary value
        with the relative path to the data file.

        Returns:

            None
        """
        if len(self.pending_output) != len(self.pending_input):
            util.logger.warning('the input and output have mismatched length')

        count = len(self.pending_output)

        if count > 0:
            proc = self._row_preprocessor

            self.pending_output = [
                self.munge(self.output_index + i, proc(row))
                for i, row in enumerate(self.pending_output)
            ]

            self.pending_input = [
                self.munge(self.output_index + i, proc(row))
                for i, row in enumerate(self.pending_input)
            ]

            self._write_root()
            self.output_index += len(self.pending_output) - 1
            self.clear()

    @contextmanager
    @util.hide_in_traceback
    def context(self, *args, **kws):
        """calls `self.new_row(*args, **kws); self.write()` on context exit.

        This is meant as a convenience for defining execution behavior in
        table inputs for Racks.
        """

        try:
            yield self
        finally:
            self.new_row(*args, **kws)
            self.write()

    def _write_root(self):
        """Write pending data. This must be implemented by inheriting classes.

        Returns:
            None
        """
        raise NotImplementedError

    def clear(self):
        """Remove any queued data that has been added by append."""
        self.pending_output = []
        self.pending_input = []

    def set_relational_file_format(self, format: str):
        """Set the format to use for relational data files.

        Arguments:
            format: one of 'csv', 'json', 'feather', or 'pickle'
        """
        warnings.warn(
            """set_nonscalar_file_type is deprecated; set when creating
                         the database object instead with the nonscalar_output flag"""
        )

        if format not in ('csv', 'json', 'feather', 'pickle', 'db'):
            raise Exception(f'relational file data format {format} not supported')
        self.munge.nonscalar_file_type = format

    def open(self):
        """Open the file or database connection.
        This is an abstract base method (to be overridden by inheriting classes)

        Returns:
            None
        """

        if self.path is None:
            raise TypeError('cannot open dB while path is None')

        # introspect self so to hook in self.host and self.munge
        if self not in self.aggregator.name_map:
            self.aggregator.update_name_map({self: None}, fallback_names={self: 'db'})

        self.observe_paramattr(self.munge, never=attr.get_class_attrs(self.munge))
        self.observe_paramattr(self.host, always=['time', 'log'])

        # configure strings in relational data files that depend on how 'self' is
        # named in introspection
        log_key = self.aggregator.name_attr_field(self.host, 'log')
        if log_key not in self.munge.force_relational:
            self.munge.force_relational = list(self.munge.force_relational) + [log_key]

        # Do some checks on the relative data directory before we consider overwriting
        # the root db.
        self.aggregator.enable()

        self.clear()

        self.output_index = 0
        self._logger.debug(f'{self} is open')
        return self

    def close(self):
        self.aggregator.disable()
        # try:
        self.write()

        if self.output_index > 0:
            if self.host.isopen:
                self.munge.save_metadata(
                    self.aggregator.get_metadata(self.munge.format_metadata_value)
                )
            else:
                self.munge._logger.warning('not saving metadata due to exception')


class CSVLogger(ParamAttrLogger):
    """Manage logging of experimental data and methods into CSV files.

    Explicit save methods are exposed for arbitrary custom data. Automatic logging is performed for

    #. Access to parameters of :class:`labbench.Device` objects that are defined with :mod:`labbench.paramattr.value`,
     :mod:`labbench.paramattr.property`, and :mod:`labbench.paramattr.method`.
    #. Function calls to methods of :class:`labbench.Rack`
    #. Metadata

    Arguments:
        path: Base path to use for the root database
        overwrite: Whether to overwrite the root database if it exists (otherwise, append)
        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data
        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
    """

    ROOT_FILE_NAME = OUTPUT_FILE_NAME = 'outputs.csv'
    INPUT_FILE_NAME = 'inputs.csv'
    output_index = 0
    tables = {}

    nonscalar_file_type = 'csv'

    def open(self):
        """Instead of calling `open` directly, consider using
        `with` statements to guarantee proper disconnection
        if there is an error. For example, the following
        sets up a connected instance::

            with CSVLogger('my.csv') as db:
                ### do the data acquisition here
                pass

        would instantiate a `CSVLogger` instance, and also guarantee
        a final attempt to write unwritten data is written, and that
        the file is closed when exiting the `with` block, even if there
        is an exception.
        """

        # touch pandas to force lazy loading
        pd.__version__

        self.tables = {}

        # checking for existing files comes after this
        self.path.mkdir(parents=True, exist_ok=True)

        for file_name in self.OUTPUT_FILE_NAME, self.INPUT_FILE_NAME:
            file_path = self.path / file_name
            try:
                # test access by starting the root table
                file_path.touch(exist_ok=self._append)
            except FileExistsError:
                raise OSError(
                    f"root table already exists at '{file_path}', while append=False"
                )

            if self._append and file_path.stat().st_size > 0:
                # there's something here and we plan to append
                self.tables[file_name] = pd.read_csv(file_path, nrows=1)
                self.tables[file_name].index.name = self.INDEX_LABEL
                self.output_index = len(self.tables[file_name])
            else:
                self.tables[file_name] = None
                self.output_index = 0

    def _write_root(self):
        """Write queued rows of data to csv. This is called automatically on :func:`close`, or when
        exiting a `with` block.

        If the class was created with overwrite=True, then the first call to _write_root() will overwrite
        the preexisting file; subsequent calls append.
        """

        def append_csv(path: Path, df: pd.DataFrame):
            if len(df) == 0:
                return

            isfirst = self.tables.get(path.name, None) is None
            pending = pd.DataFrame(df)
            pending.index.name = self.INDEX_LABEL
            pending.index = pending.index + self.output_index

            if isfirst:
                self.tables[path.name] = pending
            else:
                self.tables[path.name] = pd.concat((self.tables[path.name], pending))
                self.tables[path.name] = self.tables[path.name].loc[pending.index[0] :]
            self.tables[path.name].sort_index(inplace=True)

            self.path.mkdir(exist_ok=True, parents=True)

            self.tables[path.name].to_csv(
                path, mode='a', header=isfirst, index=False, encoding='utf-8'
            )

        if len(self.pending_input) > 0:
            append_csv(self.path / self.INPUT_FILE_NAME, self.pending_input)
        if len(self.pending_output) > 0:
            append_csv(self.path / self.OUTPUT_FILE_NAME, self.pending_output)

            output_df = self.tables[(self.path / self.OUTPUT_FILE_NAME).name]
            self.output_index = output_df.index[-1]


class SQLiteLogger(ParamAttrLogger):
    """Store data and property traits to disk into an an sqlite database.

    This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path: Base path to use for the root database
        overwrite: Whether to overwrite the root database if it exists (otherwise, append)
        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data
        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
    """

    INDEX_LABEL = 'id'  # Don't change this or sqlite breaks :(
    ROOT_FILE_NAME = 'root.db'
    OUTPUT_TABLE_NAME = 'output'
    _columns = None
    inprogress = {}
    committed = {}
    output_index = 0
    _engine = None

    def open(self):
        """Instead of calling `open` directly, consider using
        `with` statements to guarantee proper disconnection
        if there is an error. For example, the following
        sets up a connected instance::

            with SQLiteLogger('my.db') as db:
                ### do the data acquisition here
                pass

        would instantiate a `SQLiteLogger` instance, and also guarantee
        a final attempt to write unwritten data is written, and that
        the file is closed when exiting the `with` block, even if there
        is an exception.
        """

        if self.path.exists():
            root = str(self.path.absolute())
            if not self.path.is_dir():
                raise OSError(
                    f"the root data directory path '{root}' already exists, and is not a directory."
                )

            try:
                # test writes
                with open(self.path / '_', 'wb'):
                    pass
                (self.path / '_').unlink()
                ex = None
            except OSError as e:
                ex = e
            if ex:
                raise ex

        #        if not self.path.lower().endswith('.db'):
        #            self.path += '.db'
        os.makedirs(self.path, exist_ok=True)
        path = os.path.join(self.path, self.ROOT_FILE_NAME)

        # Create an engine via sqlalchemy
        self._engine = sqlalchemy.create_engine(f'sqlite:///{path}')

        if os.path.exists(path):
            if self._append:
                # read the last row to 1) list the columns and 2) get index
                query = f'select * from {self.OUTPUT_TABLE_NAME} order by {self.INDEX_LABEL} desc limit 1'

                df = pd.read_sql_query(query, self._engine, index_col=self.INDEX_LABEL)
                self._columns = df.columns
                self.output_index = df.index[-1] + 1
            else:
                raise OSError(
                    f"root table already exists at '{path}', but append=False"
                )
        else:
            self._columns = None
        self.inprogress = {}
        self.committed = {}

    def close(self):
        try:
            self.write()
            self._write_root()
        finally:
            if self._engine is not None:
                self._engine.dispose()

    def _write_root(self):
        """Write queued rows of data to the database. This also is called automatically on :func:`close`, or when
        exiting a `with` block.

        If the class was created with overwrite=True, then the first call to _write_root() will overwrite
        the preexisting file; subsequent calls append.
        """
        if len(self.pending_output) == 0:
            return

        # Put together a DataFrame from self.pending that is guaranteed
        # to include the columns defined in self._columns
        pending = pd.DataFrame(self.pending_output)
        pending.index = pending.index + self.output_index
        blank = pd.DataFrame(columns=self._columns)

        # Check for new columns, and insert into the database
        if self._columns is None:
            new_columns = []
        else:
            trait_cols = [c.lower() for c in pending.columns]
            blank_cols = [c.lower() for c in blank.columns]
            new_columns = list(set(trait_cols).difference(blank_cols))

        for c in new_columns:
            self._logger.debug(f"inserting new column '{c}'")
            column_type = self._sql_type_name(pending[c])
            with self._engine.connect() as con:
                query = f'alter table {self.OUTPUT_TABLE_NAME} add column {c} {column_type} default NULL'
                con.execute(query)

        # Form the new database row
        df = pd.concat([blank, pending], sort=True)
        df.sort_index(inplace=True)
        self._columns = df.columns

        from sqlalchemy.exc import ArgumentError

        # Append to db
        try:
            df.to_sql(
                self.OUTPUT_TABLE_NAME,
                self._engine,
                if_exists='append',
                index=True,
                index_label=self.INDEX_LABEL,
            )
        except ArgumentError:
            self._logger.error(f'failed to convert index label {self.INDEX_LABEL}')
            raise ArgumentError
        except BaseException:
            raise

        self.output_index += 1

    def key(self, name, attr):
        """The key determines the SQL column name. df.to_sql does not seem
        to support column names that include spaces
        """
        key_name = name.replace(' ', '_').replace('.', '_')
        return f'{key_name}_{attr}'

    def _sql_type_name(self, col):
        col_type = self._get_notnull_col_dtype(col)
        if col_type == 'timedelta64':
            warnings.warn(
                "the 'timedelta' type is not supported, and will be \
                           written as integer values (ns frequency) to the\
                           database."
            )
            col_type = 'integer'

        elif col_type == 'datetime64':
            col_type = 'datetime'

        elif col_type == 'empty':
            col_type = 'string'

        elif col_type == 'complex':
            raise ValueError('Complex datatypes not supported')

        if col_type not in pd.io.sql._SQL_TYPES:
            col_type = 'string'

        return pd.io.sql._SQL_TYPES[col_type]

    def _get_notnull_col_dtype(self, col):
        """
        Infer datatype of the Series col.  In case the dtype of col is 'object'
        and it contains NA values, this infers the datatype of the not-NA
        values.  Needed for inserting typed data containing NULLs, GH8778.
        """
        col_for_inference = col
        if col.dtype == 'object':
            notnulldata = col[~pd.isnull(col)]
            if len(notnulldata):
                col_for_inference = notnulldata

        return pd.api.types.infer_dtype(col_for_inference)


def to_feather(data, path):
    """
    Write a dataframe to a feather file on disk. Any index will be moved
    to a column, index and column name metadata will be removed,
    and columns names will be changed to a string.

    Arguments:
        data: dataframe to write to disk
        path: path to file to write
    Returns:
        None

    """
    iname, data.index.name = data.index.name, None
    cname, data.columns.name = data.columns.name, None

    try:
        if not (
            data.index.is_monotonic
            and data.index[0] == 0
            and data.index[-1] == data.shape[0] - 1
        ):
            data = data.reset_index()
        data.columns = np.array(data.columns).astype(np.str)
        data.to_feather(path)
    finally:
        data.index.name = iname
        data.columns.name = cname


def read_sqlite(
    path,
    table_name=SQLiteLogger.OUTPUT_TABLE_NAME,
    columns=None,
    nrows=None,
    index_col=ParamAttrLogger.INDEX_LABEL,
):
    """Wrapper to that uses pandas.read_sql_table to load a table from an sqlite database at the specified path.

    Arguments:
        path: sqlite database path
        table_name: name of table in the sqlite database
        columns: columns to query and return, or None (default) to return all columns
        nrows: number of rows of data to read, or None (default) to return all rows
        index_col: the name of the column to use as the index
    Returns:
        pandas.DataFrame instance containing data loaded from `path`
    """
    engine = sqlalchemy.create_engine(f'sqlite:///{path}')
    df = pd.read_sql_table(table_name, engine, index_col=index_col, columns=columns)
    engine.dispose()
    if nrows is not None:
        df = df.iloc[:nrows]
    return df


def read(
    path_or_buf: str,
    columns: list[str] = None,
    nrows: int = None,
    format: str = 'auto',
    **kws,
):
    """Read tabular data from a file in one of various formats
    using pandas.

    Arguments:
        path: path to the  data file.
        columns: a column or iterable of multiple columns to return from the data file, or None (the default) to return all columns
        nrows: number of rows to read at the beginning of the table, or None (the default) to read all rows
        format: data file format, one of ['pickle','feather','csv','json','csv'], or 'auto' (the default) to guess from the file extension
        kws: additional keyword arguments to pass to the pandas read_<ext> function matching the file extension
    Returns:
        pandas.DataFrame instance containing data read from file
    """

    reader_guess = {
        'p': pd.read_pickle,
        'pickle': pd.read_pickle,
        'db': read_sqlite,
        'sqlite': read_sqlite,
        'json': pd.read_json,
        'csv': pd.read_csv,
    }

    from pyarrow import feather

    try:
        reader_guess.update(
            {'f': feather.read_feather, 'feather': feather.read_feather}
        )
    except BaseException:
        warnings.warn(
            'feather format is not available in this pandas installation, and will not be supported in labbench'
        )

    if isinstance(path_or_buf, (str, Path)):
        path_or_buf = str(path_or_buf)
        if os.path.getsize(path_or_buf) == 0:
            raise OSError('file is empty')

        if format == 'auto':
            format = os.path.splitext(path_or_buf)[-1][1:]
    else:
        if format == 'auto':
            raise ValueError(
                'can only guess format for string path - specify extension'
            )

    try:
        reader = reader_guess[format]
    except KeyError as e:
        raise Exception(
            f"couldn't guess a reader from extension of file {path_or_buf}"
        ) from e

    if reader == read_sqlite:
        return reader(path_or_buf, columns=columns, nrows=nrows, **kws)
    elif reader == pd.read_csv:
        return reader(path_or_buf, usecols=columns, nrows=nrows, **kws)
    else:
        if columns is None:
            columns = slice(None, None)
        return reader(path_or_buf, **kws)[columns].iloc[:nrows]


class MungeTarReader:
    tarnames = 'data.tar', 'data.tar.gz', 'data.tar.bz2', 'data.tar.lz4'

    def __init__(self, path, tarname='data.tar'):
        self.tarfile = tarfile.open(os.path.join(path, tarname), 'r')

    def __call__(self, key, *args, **kws):
        key = key.replace('\\\\', '\\')

        for k in key, key.replace('\\', '/').replace('//', '/'):
            try:
                ext = os.path.splitext(key)[1][1:]
                return read(self.tarfile.extractfile(k), format=ext, *args, **kws)
            except KeyError as e:
                ex = e
                continue
            else:
                break
        else:
            raise ex

    def __del__(self):
        self.tarfile.close()


class MungeDirectoryReader:
    def __init__(self, path):
        self.path = Path(path)

    def __call__(self, key, *args, **kws):
        return read(os.path.join(self.path, key), *args, **kws)


class MungeReader:
    """Guess the type of munging performed on the relational data, and return
    a reader suited to loading that file.

    """

    """ TODO: Make this smarter, perhaps by trying to read an entry from
        the root database
    """

    def __new__(cls, path):
        dirname = os.path.dirname(path)

        for n in MungeTarReader.tarnames:
            if os.path.exists(os.path.join(dirname, n)):
                return MungeTarReader(dirname, tarname=n)
        return MungeDirectoryReader(dirname)


def read_relational(
    path: Union[str, Path],
    expand_col: str,
    root_cols: Union[list[str], None] = None,
    target_cols: Union[list[str], None] = None,
    root_nrows: Union[int, None] = None,
    root_format: str = 'auto',
    prepend_column_name: bool = True,
) -> DataFrameType:
    """Flatten a relational database table by loading the table located each row of
    `root[expand_col]`. The value of each column in this row
    is copied to the loaded table. The columns in the resulting table generated
    on each row are downselected
    according to `root_cols` and `target_cols`. Each of the resulting tables
    is concatenated and returned.

    The expanded dataframe may be very large, making downselecting a practical
    necessity in some scenarios.

    Arguments:
        path: file location of the root data table
        expand_col: name of the columns of the root data file to expand with relational data
        root_cols: the root columns to include in the expanded dataframe, or None (the default) pass all columns from `root`
        target_cols: the root columns to include in the expanded dataframe, or None (the default) to pass all columns loaded from each root[expand_col]
        prepend_column_name: whether to prepend the name of the expanded column from the root table

    Returns:
        the expanded dataframe

    """
    # if not isinstance(root, (pd.DataFrame,pd.Series)):
    #     raise ValueError('expected root to be a DataFrame instance, but it is {} instead'\
    #                      .format(repr(type(root))))
    if not isinstance(expand_col, str):
        raise ValueError(f'expand_col must a str, not {type(expand_col)}')

    if root_cols is not None:
        root_cols = list(root_cols) + [expand_col]
    root = read(path, columns=root_cols, nrows=root_nrows, format=root_format)

    if root_cols is not None:
        root_cols = ['root_index'] + root_cols
    root = root.reset_index(names='root_index')
    reader = MungeReader(path)

    def generate():
        for i, row in root.iterrows():
            if row[expand_col] is None or len(row[expand_col]) == 0:
                continue

            sub = reader(row[expand_col], columns=target_cols)

            if prepend_column_name:
                prepend = expand_col + '_'
            else:
                prepend = ''

            # Rename columns
            sub.columns = [prepend + c for c in sub.columns]
            sub[prepend + 'id'] = sub.index

            # # Downselect root columns
            # if root_cols is not None:
            #     row = row[root_cols]

            # Add in columns from the root
            for c, v in row.items():
                sub[c] = v

            yield sub

    return pd.concat(generate(), ignore_index=True, sort=True)



# testing Device objects implemented as a simple dict store for closed-loop get/set testing of
# method and property attributes

import typing
from collections import defaultdict
from typing import Any, Union

from .. import Device, Undefined
from .. import paramattr as attr

__all__ = ['PowerSensor', 'Oscilloscope', 'SignalGenerator']

T = typing.TypeVar('T')


class key_adapter(attr.visa_keying):
    def __init__(self, *, defaults: dict[str, Any] = {}, remap: dict[Any, str] = {}):
        super().__init__(remap=remap)
        self.defaults = defaults

    def get_kwarg_names(self, key_def: Union[str, tuple[str]]) -> tuple[str]:
        if isinstance(key_def, str) or len(key_def) == 1:
            return super().get_kwarg_names(key_def)
        else:
            return tuple(key_def[1:])

    def get(
        self,
        owner: attr.HasParamAttrs,
        key_def: Union[str, list[str]],
        paramattr: attr.ParamAttr = None,
        arguments: dict[str, Any] = {},
    ):
        """queries a parameter named `scpi_key` by sending an SCPI message string.

        The command message string is formatted as f'{scpi_key}?'.
        This is automatically called in wrapper objects on accesses to property traits that
        defined with 'key=' (which then also cast to a pythonic type).

        Arguments:
            key (str): the name of the parameter to set
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)

        Returns:
            response (str)
        """
        store_name = owner.backend.get_store_key(paramattr)
        backend_key = owner.backend.get_backend_key(owner, paramattr, arguments)

        default = self.defaults.get(store_name, paramattr._type())
        return owner.backend.get(backend_key, default)

    def set(
        self,
        owner: attr.HasParamAttrs,
        key_def: str,
        value,
        attr: attr.ParamAttr = None,
        arguments: dict[str, Any] = {},
    ):
        """writes an attribute

        The command message string is formatted as f'{scpi_key} {value}'. This
        This is automatically called on assignment to property traits that
        are defined with 'key='.

        Arguments:
            scpi_key (str): the name of the parameter to set
            value (str): value to assign
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)
        """
        backend_key = owner.backend.get_backend_key(owner, attr, arguments)
        owner.backend.set(backend_key, value)


class TestStore:
    def __init__(self):
        self.values = {}
        self.clear_counts()

    def get(self, key, default):
        self.get_count[key] += 1
        return self.values.setdefault(key, default)

    def set(self, key, value):
        self.set_count[key] += 1
        self.values[key] = value

    def clear_counts(self):
        self.get_count = defaultdict(int)
        self.set_count = defaultdict(int)
        self.notifications = []

    def get_store_key(self, attr_def: attr.ParamAttr) -> str:
        if isinstance(attr_def, attr.value.Value) or attr_def.key is Undefined:
            return attr_def.name
        elif isinstance(attr_def.key, str):
            return attr_def.key
        else:
            return attr_def.key[0]

    def get_backend_key(self, owner, attr_def, arguments={}):
        store_name = self.get_store_key(attr_def)
        # print('***', attr_def.get_key_arguments(type(owner)), arguments)

        if isinstance(attr_def, attr.method.Method):
            required_args = attr_def.get_kwarg_names()

            missing_args = set(required_args) - set(arguments)
            if len(missing_args) > 0:
                raise ValueError(f'missing required argument(s): {missing_args}')

        if len(arguments) == 0:
            return store_name
        else:
            return (store_name, frozenset(arguments.items()))

    def notification_handler(self, msg: dict):
        self.notifications.append(msg)


@key_adapter()
class StoreTestDevice(Device):
    resource: str = attr.value.str(None, kw_only=False, help='address')

    def open(self):
        self.backend = TestStore()
        attr.observe(self, self.backend.notification_handler)

    @classmethod
    def get_paramattr_arguments(
        cls, attr_name
    ) -> list[attr.method_kwarg.MethodKeywordArgument]:
        attr = getattr(cls, attr_name)
        return attr.get_key_arguments(cls)

    @classmethod
    def get_attr_defs(cls) -> dict[str, attr.ParamAttr]:
        return cls._attr_defs.attrs

    @classmethod
    def get_method_names(cls) -> list[str]:
        return cls._attr_defs.method_names()

    @classmethod
    def get_property_names(cls) -> list[str]:
        return set(cls._attr_defs.method_names()) - set(dir(Device))


class PowerSensor(StoreTestDevice):
    RATES = 'NORM', 'DOUB', 'FAST'

    # SCPI string keys and bounds on the parameter values,
    # taken from the instrument programming manual
    initiate_continuous = attr.property.bool(
        key='INIT:CONT', help='trigger continuously if True'
    )
    trigger_count = attr.property.int(
        key='TRIG:COUN', min=1, max=200, help='acquisition count', label='samples'
    )
    measurement_rate = attr.property.str(
        key='SENS:MRAT',
        only=RATES,
        case=False,
    )
    sweep_aperture = attr.property.float(
        key='SWE:APER', min=20e-6, max=200e-3, help='measurement duration', label='s'
    )
    frequency = attr.property.float(
        key='SENS:FREQ',
        min=10e6,
        max=18e9,
        step=1e-3,
        help='calibration frequency',
        label='Hz',
    )


class SpectrumAnalyzer(StoreTestDevice):
    center_frequency = attr.property.float(
        key='SENS:FREQ',
        min=10e6,
        max=18e9,
        step=1e-3,
        help='input signal center frequency',
        label='Hz',
    )
    resolution_bandwidth = attr.property.float(
        key='SENS:BW',
        min=1,
        max=40e6,
        step=1e-3,
        help='resolution bandwidth',
        label='Hz',
    )


class SignalGenerator(StoreTestDevice):
    output_enabled = attr.property.bool(
        key='OUT:ENABL', help='when True, output an RF tone'
    )
    center_frequency = attr.property.float(
        key='SENS:FREQ',
        min=10e6,
        max=18e9,
        step=1e-3,
        help='input signal center frequency',
        label='Hz',
    )
    mode = attr.property.str(key='MODE', only=['sweep', 'tone', 'iq'], case=False)


@attr.method_kwarg.int(name='channel', min=1, max=4, help='input channel')
class Oscilloscope(StoreTestDevice):
    @attr.method.float(
        min=10e6,
        max=18e9,
        step=1e-3,
        label='Hz',
        help='channel center frequency',
    )
    def center_frequency(self, *, channel):
        return self.query(f'CH{channel}:SENS:FREQ?')

    @center_frequency.setter
    def _(self, set_value, /, *, channel):
        self.write(f'CH{channel}:SENS:FREQ {set_value}')

    resolution_bandwidth = attr.method.float(
        key=('CH:SENS:BW', 'channel'),
        min=1,
        max=40e6,
        step=1e-3,
        help='channel resolution bandwidth',
        label='Hz',
        # arguments omitted deliberately for testing
    )



import builtins
import string
from typing import Any, Union
from numbers import Number

from ._bases import HasParamAttrs, KeyAdapterBase, ParamAttr, T, BoundedNumber
from ._types import Bool


class message_keying(KeyAdapterBase):
    """Base class for decorators configure wrapper access to a backend API in of a
    :class:labbench.Device` class through string `key` arguments.

    Decorated classes connect traits that are defined with the `key` keyword to trigger
    backend API calls based on the key. The implementation of the `set` and `get` methods
    determines how the key is used to generate API calls.

    Example:

        A custom implementation::

            import labbench as lb
            from labbench import paramattr as attr


            class custom_keying(attr.message_keying):
                def get(self, device: lb.Device, scpi_key: str, trait_name=None):
                    if ' ' in scpi_key:
                        key = scpi_key.replace(' ', '? ', 1)
                    else:
                        key = scpi_key + '?'
                    return device.query(key)

                def set(self, device: lb.Device, scpi_key: str, value, trait_name=None):
                    if ' ' in scpi_key:
                        key = f'{scpi_key},{value}'
                    else:
                        key = f'{scpi_key} {value}'
                    return device.write(key.rstrip())


            @custom_keying(remap={True: 'ON', False: 'OFF'})
            class CustomDevice(lb.VISADevice):
                pass

    See Also:

        * :meth:`labbench.paramattr.visa_keying`
    """

    _formatter = string.Formatter()

    def __init__(
        self,
        *,
        query_fmt=None,
        write_fmt=None,
        write_func=None,
        query_func=None,
        remap={},
    ):
        super().__init__()

        self.query_fmt = query_fmt
        self.write_fmt = write_fmt
        self.write_func = write_func
        self.query_func = query_func

        if len(remap) == 0:
            self.value_map = {}
            self.message_map = {}
            return

        # ensure str type for messages; keys can be arbitrary python type
        if not all(isinstance(v, builtins.str) for v in remap.values()):
            raise TypeError('all values in remap dict must have type str')

        self.value_map = remap

        # create the reverse mapping
        self.message_map = builtins.dict(zip(remap.values(), remap.keys()))

        # and ensure all values are unique
        if len(self.message_map) != len(self.value_map):
            raise ValueError("'remap' has duplicate values")

    def get_kwarg_names(self, s: str) -> list[str]:
        """returns an argument list based on f-string style curly-brace formatting tokens.

        Example:

            ```python
            # input
            print(key_adapter.get_kwarg_names('CH{channel}:SV:CENTERFrequency'))
            ['channel']
            ```
        """
        return [tup[1] for tup in self._formatter.parse(s) if tup[1] is not None]

    def from_message(self, msg):
        return self.message_map.get(msg, msg)

    def to_message(self, value, attr_def: ParamAttr):
        matches = tuple({value} & self.value_map.keys())
        if len(matches) == 0:
            return value

        # gymnastics caused by the python quirk that (True == 1) but not (True is 1):
        # we don't want to remap 1 using self.value_map[True]
        # TODO: change the definition of remap to fix this problem
        key_type = type(self.message_map[self.value_map[matches[0]]])
        if key_type is bool and isinstance(attr_def, BoundedNumber):
            return value
        elif (
            issubclass(key_type, Number)
            and not issubclass(key_type, bool)
            and isinstance(attr_def, Bool)
        ):
            return value

        return self.value_map[value]

    def get(
        self,
        owner: HasParamAttrs,
        scpi_key: str,
        paramattr: Union[ParamAttr[T], None],
        kwargs: dict[str, Any] = {},
    ) -> T:
        """queries a parameter named `scpi_key` by sending an SCPI message string.

        The command message string is formatted as f'{scpi_key}?'.
        This is automatically called in wrapper objects on accesses to property traits that
        defined with 'key=' (which then also cast to a pythonic type).

        Arguments:
            key (str): the name of the parameter to set
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)

        Returns:
            response (str)
        """
        if self.query_fmt is None:
            raise ValueError('query_fmt needs to be set for key get operations')
        if self.query_func is None:
            raise ValueError('query_func needs to be set for key get operations')
        query_func = getattr(owner, self.query_func)
        try:
            expanded_scpi_key = scpi_key.format(**kwargs)
        except KeyError:
            expected_kws = set(self.get_kwarg_names(scpi_key))
            missing_kws = expected_kws - set(kwargs.keys())
            raise TypeError(
                f'{paramattr._owned_name(owner)}() missing required positional argument(s) {str(missing_kws)[1:-1]}'
            )

        value_msg = query_func(self.query_fmt.format(key=expanded_scpi_key)).rstrip()
        return self.from_message(value_msg)

    def set(
        self,
        owner: HasParamAttrs,
        scpi_key: str,
        value: T,
        paramattr: Union[ParamAttr[T], None],
        kwargs: dict[str, Any] = {},
    ):
        """writes an SCPI message to set a parameter with a name key
        to `value`.

        The command message string is formatted as f'{scpi_key} {value}'. This
        This is automatically called on assignment to property traits that
        are defined with 'key='.

        Arguments:
            scpi_key (str): the name of the parameter to set
            value (str): value to assign
            name (str, None): name of the trait setting the key (or None to indicate no trait) (ignored)
        """
        if self.write_fmt is None:
            raise ValueError('write_fmt needs to be set for key set operations')
        if self.write_func is None:
            raise ValueError('write_func needs to be set for key set operations')

        value_msg = self.to_message(value, paramattr)
        expanded_scpi_key = scpi_key.format(**kwargs)
        write_func = getattr(owner, self.write_func)
        write_func(self.write_fmt.format(key=expanded_scpi_key, value=value_msg))


class visa_keying(message_keying):
    """Decorates a :class:`labbench.VISADevice` (or subclass) to configure its use of the `key` argument
    in all :mod:`labbench.paramattr.property` or :mod:`labbench.paramattr.method` descriptors.

    Example:

        Configure `MyDevice` to format SCPI queries as `'{key}?'`, and SCPI writes as f'{key} {value}'::

            import labbench as lb


            @lb.visa_keying(query_fmt='{key}?', write_fmt='{key} {value}')
            class MyDevice(lb.VISADevice):
                pass

    """

    def __init__(
        self,
        *,
        query_fmt='{key}?',
        write_fmt='{key} {value}',
        remap={},
    ):
        super().__init__(
            query_fmt=query_fmt,
            write_fmt=write_fmt,
            remap=remap,
            write_func='write',
            query_func='query',
        )



from __future__ import annotations

import contextlib
import importlib
import inspect
import logging
import os
import platform
import select
import socket
import sys
import warnings
from collections import OrderedDict
from pathlib import Path
from queue import Empty, Queue
from threading import Event, Thread
from typing import Union

import typing_extensions as typing
from typing_extensions import Literal

from . import paramattr as attr
from . import util
from ._device import Device

if typing.TYPE_CHECKING:
    import psutil
    import pyvisa
    import serial
    import subprocess as sp
else:
    psutil = util.lazy_import('psutil')
    pyvisa = util.lazy_import('pyvisa')
    serial = util.lazy_import('serial')
    sp = util.lazy_import('subprocess')


def shell_options_from_keyed_values(
    device: Device,
    skip_none=True,
    hide_false: bool = False,
    join_str: Union[Literal[False], str] = False,
    remap: dict = {},
    converter: callable = str,
) -> list[str]:
    """generate a list of command line argument strings based on
    :module:`labbench.paramattr.value` descriptors in `device`.

    Each of these descriptors defined with `key` may be treated as a command
    line option. Value descriptors are ignored when `key` is unset. The value
    for each option is determined by fetching the corresponding attribute from
    `device`.

    The returned list of strings can be used to build the `argv` needed to run
    shell commands using :class:`ShellBackend` or the `subprocess` module.

    Arguments:
        device: the device containing values to broadcast into command-line arguments
        skip_none: if True, no command-line argument string is generated for values that are unset in devices
        hide_false: if True, boolean options are treated as a flag (e.g., this triggers argument strings are omitted for False values)
        join_str: a string to use to join option (name, value) pairs, or False to generate as separate strings
        remap: a dictionary of {python_value: string_value} pairs to accommodate special cases in string conversion
        converter: function to use to convert the values to strings

    Example:
        Simple boolean options and flags:

        >>> import labbench as lb
        >>> class ShellCopy(ShellBackend):
        ...     recursive: bool = attr.value.bool(False, key='-R')
        >>> cp = ShellCopy(recursive=True)
        >>> print(shell_options_from_keyed_values(cp, hide_false=True))
        ['-R']
        >>> print(shell_options_from_keyed_values(cp, remap={True: 'yes', False: 'no'}))
        ['-R', 'yes']

    Example:
        A non-boolean option:

        >>> class DiskDuplicate(ShellBackend):
        ...     block_size: str = attr.value.str('1M', key='bs')
        >>> dd = DiskDuplicate()
        >>> dd.block_size = '1024k'
        >>> print(shell_options_from_keyed_values(dd, join_str='='))
        ['bs','1024k']
        >>> print(shell_options_from_keyed_values(dd, join_str='='))
        ['bs=1024k']
    """

    argv = []
    for name, attr_def in attr.get_class_attrs(device).items():
        if not isinstance(attr_def, attr.value.Value) or attr_def.key in (
            None,
            attr.Undefined,
        ):
            continue

        py_value = getattr(device, name)

        if not isinstance(attr_def.key, str):
            attr_desc = attr_def.__repr__(owner=device)
            raise TypeError(
                f'flags defined as keys must have type str, but {attr_desc}.key '
                f'has type {type(attr_def.key).__qualname__}'
            )

        if py_value is None:
            if skip_none:
                continue
            else:
                raise ValueError(
                    'None is an invalid flag argument value when skip_none is False'
                )

        elif attr_def._type is bool:
            if hide_false:
                str_args = [attr_def.key] if py_value else []
            elif py_value in remap:
                str_args = [attr_def.key, remap[py_value]]
            else:
                raise ValueError(
                    f'specify hide_false=True or set remap[{py_value}] to enable boolean value mapping'
                )

        else:
            py_value = remap.get(py_value, py_value)
            str_args = [attr_def.key, converter(py_value)]

        if join_str:
            str_args = [
                join_str.join(str_args),
            ]

        argv += str_args

    return argv


class ShellBackend(Device):
    """A wrapper for running shell commands.

    This is a thin wrapper around the subprocess module. Data can be captured from standard output,
    and standard error pipes, and optionally run as a background thread.

    After opening, `backend` is `None`. On a call to run(background=True), `backend`
    becomes is a subprocess instance. When EOF is reached on the executable's
    stdout, the backend resets to None.

    When `run` is called, the program runs in a subprocess.
    The output piped to the command line standard output is queued in a
    background thread. Call read_stdout() to retreive (and clear) this
    queued stdout.
    """

    background_timeout: float = attr.value.float(
        default=1,
        min=0,
        help='wait time after close before killing background processes',
        label='s',
        cache=True,
    )

    def open(self):
        """The :meth:`open` method implements opening in the
        :class:`Device` object protocol. Call the
        :meth:`execute` method when open to
        execute the binary.
        """

        def check_state_change(change={}):
            if self.running():
                raise ValueError(
                    'cannot change command line property trait traits during execution'
                )

        # a Queue for stdout
        self.backend = None

        self._stdout = Queue()
        self._stderr = Queue()

        # Monitor property trait changes
        values = set(attr.list_value_attrs(self)) - set(dir(ShellBackend))

        attr.observe(self, check_state_change, name=tuple(values))

    def run(
        self,
        *argv,
        pipe=True,
        background=False,
        check_return=True,
        raise_on_stderr=False,
        respawn=False,
        timeout=None,
    ):
        if pipe and background:
            return self._background_piped(
                *argv,
                check_return=check_return,
                raise_on_stderr=raise_on_stderr,
                timeout=timeout,
                respawn=respawn,
            )

        if respawn:
            raise ValueError('respawn argument requires pipe=True and background=True')

        if pipe and not background:
            return self._run_piped(
                *argv,
                check_return=check_return,
                raise_on_stderr=raise_on_stderr,
                timeout=timeout,
            )
        else:
            if background:
                raise ValueError('background argument requires pipe=True')
            if raise_on_stderr:
                raise ValueError('raise_on_stderr requires pipe=True')

            return self._run_simple(*argv, check_return=check_return, timeout=timeout)

    def _run_simple(
        self, *argv: list[str], check_return: bool = False, timeout: bool = None
    ):
        """Blocking execution of the command line strings specified by `argv`. If check=True, raise an exception
        on non-zero return code.

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.

        Returns:

            None?
        """
        if timeout is None:
            timeout = self.background_timeout

        return sp.run(argv, check=check_return, timeout=timeout)

    def _run_piped(
        self,
        *argv: list[str],
        check_return: bool = False,
        raise_on_stderr: bool = False,
        timeout: Union[float, None] = None,
    ) -> None:
        """Blocking execution of the specified command line, with a pipe to collect
        stdout.

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.

        Returns:
            stdout
        """
        if timeout is None:
            timeout = self.background_timeout

        path = argv[0]
        try:
            rel = os.path.relpath(path)
            if len(rel) < len(path):
                path = rel
        except ValueError:
            pass

        self._logger.debug(f"shell execute {' '.join(argv)!r}")
        cp = sp.run(
            argv, timeout=timeout, stdout=sp.PIPE, stderr=sp.PIPE, check=check_return
        )
        ret = cp.stdout

        err = cp.stderr.strip().rstrip().decode()
        if raise_on_stderr and len(err) > 0:
            raise ChildProcessError(err)

        if ret:
            lines = ret.decode().splitlines()
            show_count = min(40, len(lines))
            remaining = max(0, len(lines) - show_count)

            logger_msgs = [f'► {line}' for line in lines[: show_count // 2]]
            if remaining > 0:
                logger_msgs.append(f'…{remaining} more lines')
            for line in lines[-show_count // 2 :]:
                logger_msgs.append(f'► {line}')
            self._logger.debug('\n'.join(logger_msgs))
        return ret

    def _background_piped(
        self,
        *argv,
        check_return=False,
        raise_on_stderr=False,
        respawn=False,
        timeout=None,
    ):
        """Run the executable in the background (returning immediately while
        the executable continues running).

        Once the background process is running,

        * Retrieve standard output from the executable with self.read_stdout

        * Write to standard input self.write_stdin

        * Kill the process with self.kill

        * Check whether the process is running with self.running

        Each command line argument in argv is either
        * a string, which is passed to the binary as is, or
        * list-like sequence ('name0', 'name1', ...) that names value traits to insert as flag arguments

        Flag arguments are converted into a sequence of strings by (1) identifying the command line flag
        for each name as `self.flags[name]` (e.g., "-f"), (2) retrieving the value as getattr(self, name), and
        (3) *if* the value is not None, appending the flag to the list of arguments as appropriate.
        """

        def stdout_to_queue(fd, cmdl):
            """Thread worker to funnel stdout into a queue"""

            pid = self.backend.pid
            q = self._stdout
            for line in iter(fd.readline, ''):
                line = line.decode(errors='replace').replace('\r', '')
                if len(line) > 0:
                    q.put(line)
                else:
                    break
            self.backend = None

            # Respawn (or don't)
            if respawn and not self.__kill:
                self._logger.debug('respawning')
                self._kill_proc_tree(pid)
                spawn(cmdl)
            else:
                self._logger.debug('process ended')

        def stderr_to_exception(fd, cmdl):
            """Thread worker to raise exceptions on standard error output"""
            q = self._stderr
            for line in iter(fd.readline, ''):
                if self.backend is None:
                    break
                line = line.decode(errors='replace').replace('\r', '')
                if len(line) > 0:
                    q.put(line)
                    self._logger.debug(f'stderr {line!r}')
                #                    raise Exception(line)
                else:
                    break

        def spawn(cmdl):
            """Execute the binary in the background (nonblocking),
            while funneling its standard output to a queue in a thread.

            Arguments:
                cmd: iterable containing the binary path, then
                        each argument to be passed to the binary.

            Returns:

                None
            """
            if self.running():
                raise Exception('already running')

            if platform.system().lower() == 'windows':
                si = sp.STARTUPINFO()
                si.dwFlags |= sp.STARTF_USESHOWWINDOW
                platform_flags = {
                    'startupinfo': si,
                    'creationflags': sp.CREATE_NEW_PROCESS_GROUP,
                }
            else:
                platform_flags = {}

            proc = sp.Popen(
                list(cmdl), stdout=sp.PIPE, stderr=sp.PIPE, **platform_flags
            )

            self.backend = proc
            Thread(target=lambda: stdout_to_queue(proc.stdout, cmdl)).start()
            if raise_on_stderr:
                Thread(target=lambda: stderr_to_exception(proc.stderr, cmdl)).start()

        if not self.isopen:
            raise ConnectionError(
                f"{self} needs to be opened ({self}.open() or in a 'with' block) manage a background process"
            )

        # Generate the commandline and spawn
        self._logger.debug(f"background execute: {' '.join(argv)!r}")
        self.__kill = False
        spawn(argv)

    def read_stdout(self, wait_for=0):
        """Pop any standard output that has been queued by a background run (see `run`).
        Afterward, the queue is cleared. Starting another background run also clears the queue.

        Returns:

            stdout
        """
        result = ''

        if not self.isopen:
            raise ConnectionError(
                'open the device to read stdout from the background process'
            )

        try:
            n = 0
            while True:
                line = self._stdout.get(wait_for > 0, timeout=self.background_timeout)
                if isinstance(line, Exception):
                    raise line

                n += 1
                result += line

                if wait_for > 0 and n == wait_for:
                    break
        except Empty:
            pass

        return result

    def write_stdin(self, text):
        """Write characters to stdin if a background process is running. Raises
        Exception if no background process is running.
        """
        try:
            self.backend.stdin.write(text)
        except ConnectionError:
            raise Exception('process not running, could not write no stdin')

    def kill(self):
        """If a process is running in the background, kill it. Sends a console
        warning if no process is running.
        """
        self.__kill = True
        backend = self.backend
        if self.running():
            self._logger.debug(f'killing process {backend.pid}')
            self._kill_proc_tree(backend.pid)

    def running(self):
        """Check whether a background process is running.

        Returns:

            True if running, otherwise False
        """
        # Cache the current running one for a second in case the backend "closes"
        return self.isopen and self.backend is not None and self.backend.poll() is None

    def clear_stdout(self):
        """Clear queued standard output. Subsequent calls to `self.read_stdout()` will return
        ''.
        """
        self.read_stdout()

    def close(self):
        self.kill()

    @staticmethod
    def _kill_proc_tree(pid, including_parent=True):
        """Kill the process by pid, and any spawned child processes.
        What a dark metaphor.
        """
        try:
            parent = psutil.Process(pid)

            # Ever notice that this metaphor is very dark?
            children = parent.children(recursive=False)

            # Get the parent first to prevent respawning
            if including_parent:
                parent.kill()
                parent.wait(5)

            for child in children:
                ShellBackend._kill_proc_tree(child.pid)
        except psutil.NoSuchProcess:
            pass


class DotNetDevice(Device):
    """Base class for .NET library wrappers based on pythonnet.

    To implement a DotNetDevice subclass::

        import labbench as lb

        class MyLibraryWrapper(lb.DotNetDevice, libary=<imported python module colocated with dll>, dll_name='mylibrary.dll')
            ...

    When a DotNetDevice is instantiated, it looks to load the dll from the location of the python module
    and dll_name.

    Attributes:
        - `backend` is None after open and is available for replacement by the subclass
    """

    # these can only be set as arguments to a subclass definition
    library = attr.value.any(
        default=None, allow_none=True, sets=False
    )  # Must be a module
    dll_name = attr.value.str(default=None, allow_none=True, sets=False)

    _dlls = {}

    def open(self):
        """dynamically import a .net CLR as a python module at self.dll"""
        library = type(self).library.default
        dll_name = type(self).dll_name.default
        dll_path = Path(library.__path__[0]) / dll_name

        try:
            # static linters really don't like this, since it's created dynamically
            import clr
        except ImportError:
            raise ImportError('pythonnet module is required to use dotnet drivers')

        # base dotnet libraries needed to identify what we're working with
        clr.setPreload(False)
        clr.AddReference('System.Reflection')

        # more frustration for static linters
        import System
        from System.Reflection import Assembly

        try:
            contents = importlib.util.find_spec(library.__package__).loader.get_data(
                str(dll_path)
            )
        except BaseException:
            with open(dll_path, 'rb') as f:
                contents = f.read()

        # binary file contents
        contents = importlib.util.find_spec(library.__package__).loader.get_data(
            str(dll_path)
        )

        # dump that into dotnet
        Assembly.Load(System.Array[System.Byte](contents))

        # do the actual import
        self.dll = importlib.import_module(dll_path.stem)


@attr.message_keying(write_fmt='{key} {value}', write_func='write')
class LabviewSocketInterface(Device):
    """Base class demonstrating simple sockets-based control of a LabView VI.

    Keyed get/set with attr.property are implemented by simple ' command value'.
    Subclasses can therefore implement support for commands in
    specific labview VI similar to VISA commands by
    assigning the commands implemented in the corresponding labview VI.

    Attributes:
        - backend: connection object mapping {'rx': rxsock, 'tx': txsock}
    """

    resource: str = attr.value.NetworkAddress(
        default='127.0.0.1',
        accept_port=False,
        kw_only=False,
        help='LabView VI host address',
    )
    tx_port: int = attr.value.int(
        default=61551, help='TX port to send to the LabView VI'
    )
    rx_port: int = attr.value.int(
        default=61552, help='TX port to send to the LabView VI'
    )
    delay: float = attr.value.float(
        default=1, min=0, help='time to wait after each property trait write or query'
    )
    timeout: float = attr.value.float(
        default=2, min=0, help='maximum wait replies before raising TimeoutError'
    )
    rx_buffer_size: int = attr.value.int(default=1024, min=1)

    def open(self):
        self.backend = dict(
            tx=socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
            rx=socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
        )

        self.backend['rx'].bind((self.resource, self.rx_port))
        self.backend['rx'].settimeout(self.timeout)
        self.clear()

    def close(self):
        for sock in list(self.backend.values()):
            try:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except BaseException:
                self._logger.error('could not close socket ', repr(sock))

    def write(self, msg):
        """Send a string over the tx socket."""
        self._logger.debug(f'write {msg!r}')
        self.backend['tx'].sendto(msg, (self.resource, self.tx_port))
        util.sleep(self.delay)

    def read(self, convert_func=None):
        """Receive from the rx socket until `self.rx_buffer_size` samples
        are received or timeout happens after `self.timeout` seconds.

        Optionally, apply the conversion function to the value after
        it is received.
        """
        rx, addr = self.backend['rx'].recvfrom(self.rx_buffer_size)
        if addr is None:
            raise Exception('received no data')
        rx_disp = rx[: min(80, len(rx))] + ('...' if len(rx) > 80 else '')
        self._logger.debug(f'read {rx_disp!r}')

        key, value = rx.rsplit(' ', 1)
        key = key.split(':', 1)[1].lstrip()
        if convert_func is not None:
            value = convert_func(value)
        return {key: value}

    def clear(self):
        """Clear any data present in the read socket buffer."""
        while True:
            inputready, o, e = select.select([self.backend['rx']], [], [], 0.0)
            if len(inputready) == 0:
                break
            for s in inputready:
                try:
                    s.recv(1)
                except BaseException:
                    continue


class SerialDevice(Device):
    """Base class for wrappers that communicate via pyserial.

    This implementation is very sparse because there is in general
    no messaging string format for serial devices.

    Attributes:
        - backend (serial.Serial): control object, after open
    """

    resource: str = attr.value.str(
        cache=True,
        kw_only=False,
        help='platform-dependent serial port address or URL',
    )

    # Connection value traits
    timeout: float = attr.value.float(
        default=None,
        min=0,
        label='s',
        help='max wait time on reads before raising TimeoutError',
    )
    write_timeout: float = attr.value.float(
        default=None,
        min=0,
        label='s',
        help='max wait time on writes before raising TimeoutError.',
    )
    baud_rate: int = attr.value.int(
        default=9600,
        min=1,
        label='bytes/s',
        help='data rate of the physical serial connection.',
    )
    parity: bytes = attr.value.str(
        default=serial.PARITY_NONE,
        only=tuple(serial.PARITY_NAMES.keys()),
        help='parity in the physical serial connection.',
    )
    stopbits: float = attr.value.float(
        default=None,
        only=[1, 1.5, 2],
        label='bits',
    )
    xonxoff: bool = attr.value.bool(
        default=False, help='whether to enable software flow control on open'
    )
    rtscts: bool = attr.value.bool(
        default=False,
        allow_none=True,
        help='whether to enable hardware (RTS/CTS) flow control on open',
    )
    dsrdtr: bool = attr.value.bool(
        default=False,
        allow_none=True,
        help='whether to enable hardware (DSR/DTR) flow control on open',
    )
    bytesize: int = attr.value.int(
        default=8, allow_none=True, only=(5, 6, 7, 8), label='bits'
    )

    # Overload methods as needed to implement the Device object protocol
    def open(self):
        """Connect to the serial device with the VISA resource string defined
        in self.resource
        """
        keys = 'timeout', 'parity', 'stopbits', 'xonxoff', 'rtscts', 'dsrdtr'
        with attr.hold_attr_notifications(self):
            params = {k: getattr(self, k) for k in keys}
            params = {k: v for k, v in params.items() if v is not None}
        self.backend = serial.serial_for_url(self.resource, self.baud_rate, **params)
        self._logger.debug('opened')

    def close(self):
        """Disconnect the serial instrument"""
        self.backend.close()
        self._logger.debug('closed')

    @classmethod
    def from_hwid(cls, hwid=None, *args, **connection_params) -> SerialDevice:
        """Instantiate a new SerialDevice from a windows `hwid' string instead
        of a comport resource. A hwid string in windows might look something
        like:

        r'PCI\\VEN_8086&DEV_9D3D&SUBSYS_06DC1028&REV_21\\3&11583659&1&B3'
        """

        usb_map = cls._map_serial_hwid_to_port()
        if hwid not in usb_map:
            raise Exception(f'Cannot find serial port with hwid {hwid!r}')
        return cls(usb_map[hwid], *args, **connection_params)

    @classmethod
    def from_url(cls, url, **kws) -> SerialDevice:
        defaults = dict(
            baudrate=None,
            bytesize=None,
            parity=None,
            stopbits=None,
            timeout=None,
            xonxoff=None,
            rtscts=None,
            write_timeout=None,
            dsrdtr=None,
            inter_byte_timeout=None,
            exclusive=None,
        )
        kws = dict(defaults, **kws)
        return cls(url, **kws)

    @staticmethod
    def list_ports(hwid=None):
        """List USB serial devices on the computer

        Returns:
            list of port resource information
        """
        from serial.tools import list_ports

        ports = [
            (port.device, {'hwid': port.hwid, 'description': port.description})
            for port in list_ports.comports()
        ]
        ports = OrderedDict(ports)

        if hwid is not None:
            ports = [
                (port, meta) for port, meta in list(ports.items()) if meta['id'] == hwid
            ]

        return dict(ports)

    @staticmethod
    def _map_serial_hwid_to_label() -> dict[str, str]:
        """Map of the comports and their names.

        Returns:
            mapping {<serial port resource>: <serial port number>}
        """
        from serial.tools import list_ports

        return {port.hwid: port.name for port in list_ports.grep('')}

    @staticmethod
    def _map_serial_hwid_to_port():
        """Map of the comports and their names.

        Returns:
            mapping {<comport name>: <comport ID>}
        """
        from serial.tools import list_ports

        return OrderedDict([(port[2], port[0]) for port in list_ports.comports()])


class SerialLoggingDevice(SerialDevice):
    """Manage connection, acquisition, and data retreival on a device
    that streams logs over serial in a background thread.
    maintaining their own threads, and blocking during setup or stop
    command execution.

    Listener objects must implement an attach method with one argument
    consisting of the queue that the device manager uses to push data
    from the serial port.
    """

    poll_rate: float = attr.value.float(
        default=0.1, min=0, help='Data retreival rate from the device (in seconds)'
    )
    stop_timeout: float = attr.value.float(
        default=0.5, min=0, help='delay after `stop` before terminating run thread'
    )
    max_queue_size: int = attr.value.int(
        default=100000, min=1, help='bytes to allocate in the data retreival buffer'
    )

    def start(self):
        """Start a background thread that acquires log data into a queue.

        Returns:
            None
        """

        q = self._stdout = Queue()
        stop_event = self._stop_requested = Event()
        finish_event = self._finished = Event()

        def accumulate():
            timeout, self.backend.timeout = self.backend.timeout, 0
            self._logger.debug(f'{self!r}: started log acquisition')

            try:
                while self.isopen:
                    buf = self.backend.read_all()
                    if len(buf) > 0:
                        q.put(buf)

                    if stop_event.wait(self.poll_rate) is not True:
                        # wait to check until here to guarantee >= 1 read
                        break
            except (ConnectionError, serial.serialutil.PortNotOpenError):
                # swallow .close() race condition
                stop_event.set()
            except serial.SerialException:
                stop_event.set()
                raise
            finally:
                finish_event.set()
                self._logger.debug(f'{self!r} stopped log acquisition')
                try:
                    self.backend.timeout = timeout
                except BaseException:
                    pass

        if self.running():
            raise Exception('already running')

        Thread(target=accumulate).start()

    def stop(self):
        """Stops the logger acquisition if it is running. Returns silently otherwise.

        Returns:

            None
        """
        self._stop_requested.set()
        self._finished.wait(max(self.poll_rate, self.stop_timeout))

    def running(self):
        """Check whether the logger is running.

        Returns:
            `True` if the logger is running
        """
        return hasattr(self, '_stop') and not self._stop_requested.is_set()

    def fetch(self):
        """Retrieve and return any log data in the buffer.

        Returns:

            any bytes in the buffer
        """
        ret = b''
        try:
            while True:
                ret += self._stdout.get_nowait()
        except Empty:
            pass
        return ret

    def clear(self):
        """Throw away any log data in the buffer."""
        self.fetch()

    def close(self):
        self.stop()


class TelnetDevice(Device):
    """A general base class for communication devices via telnet.
    Unlike (for example) VISA instruments, there is no
    standardized command format like SCPI. The implementation is
    therefore limited to open and close, which open
    or close a pyserial connection object: the `backend` attribute.
    Subclasses can read or write with the backend attribute like they
    would any other telnetlib instance.

    A TelnetDevice `resource` string is an IP address. The port is specified
    by `port`. These can be set when you instantiate the TelnetDevice
    or by setting them afterward in `value traits`.

    Subclassed devices that need property trait descriptors will need
    to implement get_key and set_key methods to implement
    the property trait set and get operations (as appropriate).
    """

    # Connection value traits
    resource: str = attr.value.NetworkAddress(
        cache=True,
        kw_only=False,
        accept_port=True,
        help='server host address',
    )
    timeout: float = attr.value.float(
        default=2, min=0, label='s', help='connection timeout'
    )

    def open(self):
        """Open a telnet connection to the host defined
        by the string in self.resource
        """
        host, *port = self.resource.split(':')

        if len(port) > 0:
            port = int(port[0])
        else:
            port = 23

        # TODO: find an alternative to allow python 3.13 support
        try:
            import telnetlib
        except ImportError:
            raise ImportError('telnet support requires python < 3.13')

        self.backend = telnetlib.Telnet(self.resource, port=port, timeout=self.timeout)

    def close(self):
        """Disconnect the telnet connection"""
        self.backend.close()


_pyvisa_resource_managers = {}


@attr.visa_keying(
    query_fmt='{key}?', write_fmt='{key} {value}', remap={True: 'ON', False: 'OFF'}
)
class VISADevice(Device):
    r"""A wrapper for VISA device automation.

    This exposes `pyvisa` instrument automation capabilities in a `labbench`
    object. Automatic connection based on make and model is supported.

    Customized operation for specific instruments should be implemented in
    subclasses.

    Examples:

        Connect to a VISA device using a known resource string::

            with VISADevice('USB0::0x2A8D::0x1E01::SG56360004::INSTR') as instr:
                print(inst)

        Probe available connections and print valid `VISADevice` constructors::

            print(visa_probe_devices())

        Probe details of available connections and identity strings on the command line::

            labbench visa-probe

        Connect to instrument with serial number 'SG56360004' and query ':FETCH?' CSV::

            with VISADevice('SG56360004') as instr:
                print(inst.query_ascii_values(':FETCH?'))

    See also:
        * Pure python backend installation:
            https://pyvisa.readthedocs.io/projects/pyvisa-py/en/latest/installation.html

        * Proprietary backend installation:
            https://pyvisa.readthedocs.io/en/latest/faq/getting_nivisa.html#faq-getting-nivisa

        * Resource strings and basic configuration:
            https://pyvisa.readthedocs.io/en/latest/introduction/communication.html#getting-the-instrument-configuration-right

    Attributes:
        backend (pyvisa.Resource): instance of a pyvisa instrument object (when open)
    """

    resource: str = attr.value.str(
        default=None,
        cache=True,
        kw_only=False,
        help='VISA resource addressing string for device connection',
    )

    read_termination: str = attr.value.str(
        default='\n',
        cache=True,
        help='end-of-line string to delineate the end of ascii query replies',
    )

    write_termination: str = attr.value.str(
        default='\n', cache=True, help='end-of-line string to send after writes'
    )

    open_timeout: float = attr.value.float(
        default=None,
        allow_none=True,
        help='timeout for opening a connection to the instrument',
        label='s',
    )

    timeout: float = attr.value.float(
        default=None,
        cache=True,
        allow_none=True,
        help='message response timeout',
        label='s',
    )

    make = attr.value.str(
        default=None,
        allow_none=True,
        cache=True,
        help='device manufacturer name used to autodetect resource string',
    )

    model = attr.value.str(
        default=None,
        allow_none=True,
        cache=True,
        help='device model used to autodetect resource string',
    )

    @attr.property.str(sets=False, cache=True)
    def serial(self):
        """device-reported serial number"""
        make, model, serial, rev = _visa_parse_identity(self._identity)
        return serial

    @attr.property.str(sets=False, cache=True, help='device revision information')
    def _revision(self):
        """device-reported revision"""
        make, model, serial, rev = _visa_parse_identity(self._identity)
        return rev

    # Common VISA properties
    _identity = attr.property.str(
        key='*IDN',
        sets=False,
        cache=True,
        help='identity string reported by the instrument',
    )

    @attr.property.dict(sets=False, log=False)
    def status_byte(self):
        """instrument status decoded from '*STB?'"""
        code = int(self.query('*STB?'))

        return {
            'error queue not empty': bool(code & 0b00000100),
            'questionable state': bool(code & 0b00001000),
            'message available': bool(code & 0b00010000),
            'event status flag': bool(code & 0b00100000),
            'service request': bool(code & 0b01000000),
            'top level status summary': bool(code & 0b01000000),
            'operating': bool(code & 0b10000000),
        }

    _rm = None  # set at runtime
    _opc = False

    # Overload methods as needed to implement RemoteDevice
    def open(self):
        """opens the instrument.

        When managing device connection through a `with` context,
        this is called automatically and does not need
        to be invoked.
        """

        if type(self)._rm is None:
            visa_default_resource_manager()

        self._opc = False

        kwargs = dict(
            read_termination=self.read_termination,
            write_termination=self.write_termination,
        )

        is_valid_visa_name = _visa_valid_resource_name(self.resource)
        if is_valid_visa_name:
            # use the explicit visa resource name, if provided
            pass
        elif (self.make, self.model) != (None, None) or self.resource:
            if self.resource:
                self._logger.debug(
                    'treating resource as a serial number (pyvisa does not recognize it as a VISA name)'
                )
            # match the supplied (make, model) and/or treat self.resource as a serial number to match
            search_desc = ', '.join(
                [
                    f'{name} {getattr(self, name)!r}'
                    for name in ('make', 'model', 'resource')
                    if getattr(self, name)
                ]
            ).replace('resource', 'serial number')

            matches = visa_probe_devices(self)

            if len(matches) == 0:
                msg = (
                    f'could not open VISA device {type(self)!r}: resource not specified, '
                    f'and no devices were discovered matching {search_desc}'
                )
                raise OSError(msg)
            elif len(matches) == 1:
                self._logger.debug(f'probed resource by matching {search_desc}')
                self.resource = matches[0].resource
            else:
                msg = (
                    f'resource ambiguity: {len(matches)} VISA resources matched {search_desc}, '
                    f'disconnect {len(matches)-1} or specify explicit resource names'
                )
                raise OSError(msg)
        else:
            raise ConnectionError(
                f'must specify a pyvisa resource name, an instrument serial number, or define {type(self)!r} with default make and model'
            )

        if self.timeout is not None:
            kwargs['timeout'] = int(self.timeout * 1000)
        if self.open_timeout is not None:
            kwargs['open_timeout'] = int(self.open_timeout * 1000)

        rm = self._get_rm()
        self.backend = rm.open_resource(self.resource, **kwargs)

        if self.timeout is not None:
            self.backend.set_visa_attribute(
                pyvisa.constants.ResourceAttribute.timeout_value,
                int(self.timeout * 1000),
            )

    def close(self):
        """closes the instrument.

        When managing device connection through a `with` context,
        this is called automatically and does not need
        to be invoked.
        """
        if not self.isopen or self.backend is None:
            return

        try:
            if hasattr(self.backend.visalib, 'viGpibControlREN'):
                with contextlib.suppress(pyvisa.errors.VisaIOError):
                    self.backend.visalib.viGpibControlREN(
                        self.backend.session, pyvisa.constants.VI_GPIB_REN_ADDRESS_GTL
                    )

        except BaseException as e:
            e = str(e)
            if len(e.strip()) > 0:
                # some emulated backends raise empty errors
                self._logger.warning('unhandled close error: ' + e)

        finally:
            self.backend.close()

    def write(self, msg: str, kws: dict[str, typing.Any] = {}):
        """sends an SCPI message to the device.

        Wraps `self.backend.write`, and handles debug logging and adjustments
        when in overlap_and_block
        contexts as appropriate.

        Arguments:
            msg: the SCPI command to send

        Returns:
            None
        """

        # append the *OPC if we're in an overlap and add context
        # TODO: this implementation doesn't generalize, since not all instruments
        # support *OPC in this context
        if self._opc:
            msg = msg + ';*OPC'

        # substitute message based on remap() in self._keying
        kws = {k: self._keying.to_message(v) for k, v in kws.items()}
        msg = msg.format(**kws)

        # outbound message as truncated event log entry
        msg_out = repr(msg) if len(msg) < 1024 else f'({len(msg)} bytes)'
        self._logger.debug(f'write({msg_out})')
        self.backend.write(msg)

    def query(
        self,
        msg: str,
        timeout=None,
        remap: bool = False,
        kws: dict[str, typing.Any] = {},
    ) -> str:
        """queries the device with an SCPI message and returns its reply.

        Handles debug logging and adjustments when in overlap_and_block
        contexts as appropriate.

        Arguments:
            msg: the SCPI message to send
        """
        if timeout is not None:
            _to, self.backend.timeout = self.backend.timeout, timeout

        # substitute message based on remap() in self._keying
        kws = {k: self._keying.to_message(v) for k, v in kws.items()}
        msg = msg.format(**kws)

        # outbound message as truncated event log entry
        msg_out = repr(msg) if len(msg) < 80 else f'({len(msg)} bytes)'
        self._logger.debug(f'query({msg_out}):')

        try:
            ret = self.backend.query(msg)
        finally:
            if timeout is not None:
                self.backend.timeout = _to

        # inbound response as truncated event log entry
        msg_out = repr(ret) if len(ret) < 80 else f'({len(ret)} bytes)'
        self._logger.debug(f'    → {msg_out}')

        if remap:
            return self._keying.from_message(ret)
        else:
            return ret

    def query_ascii_values(
        self,
        msg: str,
        converter='f',
        separator=',',
        container=list,
        delay=None,
        timeout=None,
    ):
        # pre debug
        if timeout is not None:
            _to, self.backend.timeout = self.backend.timeout, timeout

        msg_out = repr(msg) if len(msg) < 80 else f'({len(msg)} bytes)'
        self._logger.debug(f'query_ascii_values({msg_out}):')

        try:
            ret = self.backend.query_ascii_values(
                msg,
                converter=converter,
                separator=separator,
                container=container,
                delay=delay,
            )
        finally:
            if timeout is not None:
                self.backend.timeout = _to

        # post debug
        if len(ret) < 80 and len(repr(ret)) < 80:
            logmsg = repr(ret)
        elif hasattr(ret, 'shape'):
            logmsg = f'({type(ret).__qualname__} with shape {ret.shape})'
        elif hasattr(ret, '__len__'):
            logmsg = f'({type(ret).__qualname__} with length {len(ret)})'
        else:
            logmsg = f'(iterable sequence of type {type(ret)})'

        self._logger.debug(f'    -> {logmsg}')

        return ret

    def wait(self):
        """sends '*WAI' to wait for all commands to complete before continuing"""
        self.write('*WAI')

    def preset(self):
        """sends '*RST' to reset the instrument to preset"""
        self.write('*RST')

    @contextlib.contextmanager
    def overlap_and_block(
        self, timeout: float = None, quiet: bool = False, query_func: callable = None
    ):
        """context manager that sends '*OPC' on entry, and performs
        a blocking '*OPC?' query on exit.

        By convention, these SCPI commands give a hint to the instrument
        that commands sent inside this block may be executed concurrently.
        The context exit then blocks until all of the commands have
        completed.

        Example::

            with inst.overlap_and_block():
                inst.write('long running command 1')
                inst.write('long running command 2')

        Arguments:
            timeout: maximum time to wait for '*OPC?' reply, or None to use `self.backend.timeout`
            quiet: Suppress timeout exceptions if this evaluates as True

        Raises:
            TimeoutError: on '*OPC?' query timeout
        """
        self._opc = True
        yield
        self._opc = False

        if query_func is not None:
            query_func(self, timeout=timeout)
        else:
            self.query('*OPC?', timeout=timeout)

    class suppress_timeout(contextlib.suppress):
        """context manager that suppresses timeout exceptions on `write` or `query`.

        Example::

            with inst.suppress_timeout():
                inst.write('long command 1')
                inst.write('long command 2')

            If the command 1 raises an exception, command 2 will not execute
            the context block is complete, and the exception from command 1
            is swallowed.
        """

        def __exit__(self, exctype, excinst, exctb):
            EXC = pyvisa.errors.VisaIOError
            CODE = pyvisa.errors.StatusCode.error_timeout

            return exctype == EXC and excinst.error_code == CODE

    def _get_rm(self):
        backend_name = self._rm

        if backend_name in ('@ivi', '@ni'):
            is_ivi = True
            # compatibility layer for changes in pyvisa 1.12
            if 'ivi' in pyvisa.highlevel.list_backends():
                backend_name = '@ivi'
            else:
                backend_name = '@ni'
        else:
            is_ivi = False

        if len(_pyvisa_resource_managers) == 0:
            visa_default_resource_manager(backend_name)

        try:
            rm = _pyvisa_resource_managers[backend_name]
        except OSError as e:
            if is_ivi:
                url = r'https://pyvisa.readthedocs.io/en/latest/faq/getting_nivisa.html#faq-getting-nivisa'
                msg = f'could not connect to resource manager - see {url}'
                e.args[0] += msg
            raise e

        return rm


def _visa_missing_pyvisapy_support() -> list[str]:
    """a list of names of resources not supported by the current pyvisa-py install"""
    missing = []

    # gpib
    try:
        warnings.filterwarnings('ignore', 'GPIB library not found')
        import gpib_ctypes

        if not gpib_ctypes.gpib._load_lib():
            missing.append('GPIB')
    except ModuleNotFoundError:
        missing.append('GPIB')

    # hislip discovery
    if not importlib.util.find_spec('zeroconf'):
        missing.append('TCPIP')

    # libusb
    try:
        import usb1

        with usb1.USBContext():
            pass
    except OSError:
        missing.append('USB')

    # 2nd check for libusb
    import usb.core

    try:
        usb.core.find()
    except usb.core.NoBackendError:
        missing.append('USB')

    return missing


def _visa_parse_identity(identity: str):
    return identity.split(',', 4)


def visa_list_resources(resourcemanager: str = None) -> list[str]:
    """autodetects and returns a list of valid VISADevice resource strings"""
    if resourcemanager is None:
        rm = VISADevice()._get_rm()
    else:
        rm = pyvisa.ResourceManager(resourcemanager)

    return rm.list_resources()


def visa_default_resource_manager(name: str = None):
    """set the pyvisa resource manager used by labbench.

    Arguments:
        name: the name of the resource manager, such as '@py', '@sim-labbench', or '@ivi'
    """
    import pyvisa

    if name is None:
        import pyvisa.ctwrapper

        libs = pyvisa.ctwrapper.IVIVisaLibrary.get_library_paths()
        if len(libs) > 0:
            name = '@ivi'
        else:
            util.logger.info(
                'using pyvisa-py backend as fallback because no @ivi is installed'
            )
            name = '@py'

    elif name == '@sim-labbench':
        from .testing import pyvisa_sim_resource as name

    if name == '@py':
        warnings.filterwarnings(
            'ignore', 'VICP resources discovery requires the zeroconf package'
        )
        warnings.filterwarnings(
            'ignore', 'TCPIP::hislip resource discovery requires the zeroconf package'
        )
        warnings.filterwarnings('ignore', 'GPIB library not found')

    if name not in _pyvisa_resource_managers:
        _pyvisa_resource_managers[name] = pyvisa.ResourceManager(name)
    VISADevice._rm = name


@util.ttl_cache(10)  # a cache of recent resource parameters
def _visa_probe_resource(
    resource: str, open_timeout, timeout, encoding: ascii
) -> VISADevice:
    device = VISADevice(resource, open_timeout=open_timeout, timeout=timeout)
    device._logger = logging.getLogger()  # suppress the normal logger for probing

    def reopen():
        device.close()
        device.open()

    @util.retry(pyvisa.errors.VisaIOError, tries=3, log=False, exception_func=reopen)
    def probe_read_termination():
        query = '*IDN?' + device.write_termination
        device.backend.write_raw(query.encode(encoding))
        identity = device.backend.read_raw().decode(encoding)

        for read_termination in ('\r\n', '\n', '\r'):
            if identity.endswith(read_termination):
                identity = identity[: -len(read_termination)]
                break
        else:
            read_termination = ''

        return identity, read_termination

    try:
        device.open()
    except TimeoutError:
        return None

    for write_term in ('\n', '\r', '\r\n'):
        device.backend.write_termination = device.write_termination = write_term

        try:
            identity, read_termination = probe_read_termination()
            try:
                make, model, serial, *rev = _visa_parse_identity(identity)
            except ValueError:
                continue

            device.read_termination = read_termination
            device.make = make
            device.model = model
            device._attr_store.cache.update(serial=serial, _revision=rev)

            break
        except pyvisa.errors.VisaIOError as ex:
            if 'VI_ERROR_TMO' not in str(ex):
                raise
        except BaseException as ex:
            device._logger.debug(
                f'visa_list_identities exception on probing identity: {ex!s}'
            )
            raise
    else:
        device = None

    if device is not None:
        device.close()

    return device


def _visa_valid_resource_name(resource: str):
    from pyvisa import rname

    if resource is None:
        return False

    try:
        rname.ResourceName.from_string(resource)
    except rname.InvalidResourceName:
        return False
    else:
        return True


def visa_probe_devices(
    target: VISADevice = None,
    skip_interfaces: list[str] = [],
    open_timeout: float = 0.5,
    timeout: float = 0.25,
) -> list[VISADevice]:
    """discover devices available for communication and their required connection settings.

    Each returned `VISADevice` is and set with a combination of `resource`,
    `read_termination`, and `write_termination` that establishes communication, and the
    `make` and `model` reported by the instrument. The `pyvisa` resource manager is set
    by the most recent call to `visa_default_resource_manager`.

    The probe mechanism is to open a connection to each available resource, and
    if successful, query the instrument identity ('*IDN?'). Discovery will fail for
    devices that do not support this message.

    Arguments:
        target: if specified, return only devices that match target.make and target.model
        skip_interfaces: do not probe interfaces that begin with these strings (case-insensitive)
        open_timeout: timeout on resource open (in s)
        timeout: timeout on identity query (in s)
    """

    def keep_interface(name):
        for iface in skip_interfaces:
            if name.lower().startswith(iface.lower()):
                return False
        return True

    def match_target(device):
        if target is None:
            return True

        if target.make is not None:
            # apply the make filter
            if device.make.lower() != target.make.lower():
                return False

        if target.model is not None:
            # apply the model filter
            if not target.model.lower().startswith(device.model.lower()):
                return False

        if target.resource is not None and not _visa_valid_resource_name(
            target.resource
        ):
            # treat the resource string as a serial number, and filter
            if target.resource.lower() != device.serial.lower():
                return False

        return True

    calls = {
        res: util.Call(_visa_probe_resource, res, open_timeout, timeout, 'ascii')
        for res in visa_list_resources()
        if keep_interface(res)
    }

    devices = util.concurrently(**calls, catch=False, flatten=False)

    if len(devices) == 0:
        return []

    if target is not None:
        if not isinstance(target, Device) and issubclass(target, Device):
            target = target()

        devices = {
            resource: device
            for resource, device in devices.items()
            if match_target(device)
        }

    return list(devices.values())


class Win32ComDevice(Device):
    """Basic support for calling win32 COM APIs.

    a dedicated background thread. Set concurrency=True to decide whether
    this thread support wrapper is applied to the dispatched Win32Com object.
    """

    concurrency = attr.value.bool(
        default=True,
        sets=False,
        help='if False, enforces locking for single-threaded access',
    )

    # The python wrappers for COM drivers still basically require that
    # threading is performed using the windows COM API. Compatibility with
    # the python GIL is not for the faint of heart. Threading support is
    # instead realized with util.ThreadSandbox, which ensures that all calls
    # to the dispatched COM object block until the previous calls are completed
    # from within.

    com_object = attr.value.str(
        default='', sets=False, help='the win32com object string'
    )  # Must be a module

    def open(self):
        """Connect to the win32 com object"""
        import win32com
        import win32com.client

        def should_sandbox(obj):
            try:
                name = win32com.__name__
                return inspect.getmodule(obj).__name__.startswith(name)
            except AttributeError:
                return False

        def factory():
            from pythoncom import CoInitialize

            CoInitialize()
            return win32com.client.Dispatch(self.com_object)

        # Oddness for win32 threadsafety
        sys.coinit_flags = 0

        if self.com_object == '':
            raise Exception('value traits.com_object needs to be set')

        if self.concurrency:
            self.backend = util.ThreadSandbox(factory, should_sandbox)
        else:
            self.backend = win32com.client.Dispatch(self.com_object)



from . import paramattr, util
from .util import (
    Call,
    concurrently,
    logger,
    retry,
    sequentially,
    show_messages,
    sleep,
    stopwatch,
    timeout_iter,
    until_timeout,
)

util.force_full_traceback(True)

from ._backends import (
    DotNetDevice,
    LabviewSocketInterface,
    SerialDevice,
    SerialLoggingDevice,
    ShellBackend,
    TelnetDevice,
    VISADevice,
    Win32ComDevice,
    shell_options_from_keyed_values,
    visa_default_resource_manager,
    visa_list_resources,
    visa_probe_devices,
)
from ._data import CSVLogger, SQLiteLogger, read, read_relational
from ._device import Device
from ._host import Email
from ._rack import (
    Rack,
    Sequence,
    find_owned_rack_by_type,
    import_as_rack,
    rack_input_table,
    rack_kwargs_skip,
    rack_kwargs_template,
)
from ._serialize import dump_rack, load_rack
from ._version import __version__
from .paramattr import Undefined

# scrub __module__ for cleaner repr() and doc
for _obj in dict(locals()).values():
    if getattr(_obj, '__module__', '').startswith('labbench.'):
        _obj.__module__ = 'labbench'
del _obj

util.force_full_traceback(False)


def _force_full_traceback(force: bool) -> None:
    """configure whether to disable traceback hiding for internal API calls inside labbench"""
    logger.warning(
        'labbench._force_full_traceback has been deprecated - use labbench.util.force_full_traceback instead'
    )
    util.force_full_traceback(True)



from .. import util
from . import _bases

import numbers
import typing
from pathlib import Path as _Path

if typing.TYPE_CHECKING:
    import validators as _val
else:
    _val = util.lazy_import('validators')


class Any(_bases.ParamAttr[typing.Any], type=object):
    """allows any value"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        return value

    @util.hide_in_traceback
    def to_pythonic(self, value):
        return value


class Int(_bases.BoundedNumber[int], type=int):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""


class Float(_bases.BoundedNumber[float], type=float):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""

    step: typing.Union[float, None] = None

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        value = super().validate(value, owner)
        if self.step is not None:
            mod = value % self.step
            if mod < self.step / 2:
                return value - mod
            else:
                return value - (mod - self.step)
        return value

    def doc_params(
        self, skip: list[str] = ['help', 'label'], as_argument: bool = False
    ) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            docs = []
            if self.step is not None:
                docs.append(f'* Step size: {self.step} {self.label}')

            return '\n'.join(docs) + '\n'


class Complex(_bases.ParamAttr[complex], type=complex):
    """accepts numerical or str values, following normal python casting procedures (with bounds checking)"""

    allow_none: bool = False


class Bool(_bases.ParamAttr[bool], type=bool):
    """accepts boolean or numeric values, or a case-insensitive match to one of ('true',b'true','false',b'false')"""

    allow_none: bool = False
    # because this becomes confusing with evaluating boolean trueness


class String(_bases.ParamAttr):
    """base class for string types, which adds support for case sensitivity arguments"""

    case: bool = True
    # allow_none: bool = True # let's not override this default

    @util.hide_in_traceback
    def contains(self, iterable, value):
        if not self.case:
            iterable = [v.lower() for v in iterable]
            value = value.lower()
        return value in iterable

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if isinstance(value, (self._type, numbers.Number)):
            return self._type(value)

        raise ValueError(
            f'cannot convert to {type(value).__qualname__} to {self._type.__qualname__}'
        )

    def doc_params(
        self, skip: list[str] = ['help', 'label'], as_argument: bool = False
    ) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            if self.case:
                return '* Case sensitive'
            else:
                return '* Case insensitive'


class Unicode(String[str], type=str):
    """accepts strings or numeric values only; convert others explicitly before assignment"""


class Bytes(String[bytes], type=bytes):
    """accepts bytes objects only - encode str (unicode) explicitly before assignment"""


class Iterable(_bases.ParamAttr):
    """accepts any iterable"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not hasattr(value, '__iter__'):
            raise ValueError(
                f"'{type(self).__qualname__}' attributes accept only iterable values"
            )
        return self._type(value)


class Dict(Iterable[dict], type=dict):
    """accepts any type of iterable value accepted by python `dict()`"""


class List(Iterable[list], type=list):
    """accepts any type of iterable value accepted by python `list()`"""


class Tuple(Iterable[tuple], type=tuple):
    """accepts any type of iterable value accepted by python `tuple()`"""

    sets: bool = False


class Path(_bases.ParamAttr[_Path], type=_Path):
    must_exist: bool = False
    """ does the path need to exist when set? """

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        path = _Path(value)

        if self.must_exist and not path.exists():
            raise OSError('no file at path {value}')

        return path

    def doc_params(
        self, skip: list[str] = ['help', 'label'], as_argument: bool = False
    ) -> str:
        if as_argument:
            return None

        else:
            # for text docs: allow subclasses to document their own params
            if self.must_exist:
                return '* Path name must exist on the host'


class NetworkAddress(Unicode):
    """a IDN-compatible network address string, such as an IP address or DNS hostname"""

    accept_port: bool = True

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        """roughly IDN-compatible address (not URL) validator"""

        def any_valid(s: str, validators: list[callable]):
            for validator in validators:
                if validator(s):
                    return True
            return False

        def validate_port_string(s):
            if not self.accept_port:
                raise ValueError(
                    f'{self} does not accept a port number (accept_port=False)'
                )

            try:
                int(s)
            except ValueError:
                return False

            return True

        if any_valid(value, (_val.ipv4, _val.ipv6, _val.domain, _val.slug)):
            return value

        if value[0] == '[':
            # maybe ipv6 with port number
            host, *port = value[1:].rsplit(']', 1)
            if len(port) > 0 and _val.ipv6(host):
                if validate_port_string(port[0][1:]):
                    return value

        # maybe an IP address with
        host, *extra = value.split(':', 1)
        if len(extra) > 0 and any_valid(host, (_val.ipv4, _val.domain, _val.slug)):
            if validate_port_string(extra[0]):
                return value

        raise ValueError('invalid host address')



from . import method, method_kwarg, property, value
from ._bases import (
    HasParamAttrs,
    KeyAdapterBase,
    ParamAttr,
    Undefined,
    adjust,
    get_class_attrs,
    hold_attr_notifications,
    observe,
    unobserve,
    list_method_attrs,
    list_property_attrs,
    list_value_attrs,
)
from ._key_adapters import message_keying, visa_keying

for _obj in dict(locals()).values():
    if getattr(_obj, '__module__', '').startswith('labbench.paramattr.'):
        _obj.__module__ = 'labbench.paramattr'
del _obj



from . import _types
from ._bases import Property


class any(Property, _types.Any):
    pass


class bool(Property, _types.Bool):
    pass


class float(Property, _types.Float):
    pass


class int(Property, _types.Int):
    pass


class complex(Property, _types.Complex):
    pass


class str(Property, _types.Unicode):
    pass


class bytes(Property, _types.Bytes):
    pass


class list(Property, _types.List):
    pass


class tuple(Property, _types.Tuple):
    pass


class dict(Property, _types.Dict):
    pass


class Path(Property, _types.Path):
    pass


class NetworkAddress(Property, _types.NetworkAddress):
    pass
